Package: httptest2
Type: Package
Title: Test Helpers for 'httr2'
Description: Testing and documenting code that communicates with remote servers
    can be painful. This package helps with writing tests for packages that
    use 'httr2'. It enables testing all of the logic
    on the R sides of the API without requiring access to the
    remote service, and it also allows recording real API responses to use as
    test fixtures. The ability to save responses and load them offline also
    enables writing vignettes and other dynamic documents that can be
    distributed without access to a live server.
Version: 1.2.2
Authors@R: c(
        person("Neal", "Richardson", role=c("aut", "cre"), email="neal.p.richardson@gmail.com", comment=c(ORCID="0009-0002-7992-3520")),
        person("Jonathan", "Keane", role="ctb", email="jkeane@gmail.com"),
        person("Maëlle", "Salmon", role="ctb", email="maelle.salmon@yahoo.se", comment=c(ORCID="0000-0002-2815-0399"))
    )
URL: https://enpiar.com/httptest2/,
        https://github.com/nealrichardson/httptest2
BugReports: https://github.com/nealrichardson/httptest2/issues
License: MIT + file LICENSE
Imports: digest, httr2 (>= 0.2.0), jsonlite, rlang, stats, testthat (>=
        3.1.5), utils
Suggests: curl, knitr, pkgload, rmarkdown, spelling, webfakes, xml2
Language: en-US
RoxygenNote: 7.3.3
Encoding: UTF-8
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2025-11-16 19:42:32 UTC; npr
Author: Neal Richardson [aut, cre] (ORCID:
    <https://orcid.org/0009-0002-7992-3520>),
  Jonathan Keane [ctb],
  Maëlle Salmon [ctb] (ORCID: <https://orcid.org/0000-0002-2815-0399>)
Maintainer: Neal Richardson <neal.p.richardson@gmail.com>
Repository: CRAN
Date/Publication: 2025-11-16 20:10:02 UTC
Built: R 4.6.0; ; 2026-04-28 18:22:04 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(cnd_body,httptest2_request)
S3method(cnd_footer,httptest2_request)
S3method(cnd_header,httptest2_request)
export(.mockPaths)
export(block_requests)
export(build_mock_url)
export(capture_requests)
export(change_state)
export(end_vignette)
export(expect_DELETE)
export(expect_GET)
export(expect_PATCH)
export(expect_POST)
export(expect_PUT)
export(expect_no_request)
export(expect_request_header)
export(find_mock_file)
export(get_current_redactor)
export(gsub_response)
export(redact_cookies)
export(redact_headers)
export(save_response)
export(set_redactor)
export(start_capturing)
export(start_vignette)
export(stop_capturing)
export(stop_mocking)
export(use_httptest2)
export(use_mock_api)
export(with_mock_api)
export(with_mock_dir)
export(with_trace)
export(within_body_text)
export(without_internet)
importFrom(digest,digest)
importFrom(httr2,req_perform)
importFrom(httr2,request)
importFrom(httr2,resp_body_string)
importFrom(httr2,resp_content_type)
importFrom(httr2,resp_status)
importFrom(httr2,response)
importFrom(jsonlite,prettify)
importFrom(rlang,abort)
importFrom(rlang,cnd_body)
importFrom(rlang,cnd_footer)
importFrom(rlang,cnd_header)
importFrom(rlang,is_string)
importFrom(stats,terms)
importFrom(testthat,expect_error)
importFrom(testthat,expect_match)
importFrom(testthat,expect_no_error)
importFrom(testthat,expect_null)
importFrom(testthat,fail)
importFrom(utils,sessionInfo)
importFrom(utils,tail)

# httptest2: Test Helpers for 'httr2'

[![Build Status](https://github.com/nealrichardson/httptest2/workflows/R-CMD-check/badge.svg)](https://github.com/nealrichardson/httptest2/actions)
[![codecov](https://codecov.io/gh/nealrichardson/httptest2/branch/main/graph/badge.svg?token=PxOGymudAS)](https://app.codecov.io/gh/nealrichardson/httptest2)

`httptest2` makes it easy to write tests for code and packages that wrap web APIs.
Testing code that communicates with remote servers can otherwise be painful: things like authentication, server state, and network flakiness can make testing seem too costly to bother with. The `httptest2` package enables you to test all of the logic on the R sides of the API in your package without requiring access to the remote service.

Importantly, it provides **contexts** that mock the network connection and tools for **recording** real requests for future offline use as fixtures, both in tests and in vignettes. The package also includes additional **expectations** to assert that HTTP requests were---or were not---made.

Using these tools, you can test that code is making the intended requests and that it handles the expected responses correctly, all without depending on a connection to a remote API. The ability to save responses and load them offline also enables you to write package vignettes and other dynamic documents that can be distributed without access to a live server.

This package is an adaptation of [`httptest`](https://enpiar.com/r/httptest/) to work with [`httr2`](https://httr2.r-lib.org/). Most features work exactly as they do in `httptest`; see the NEWS.md for the initial release for a summary of what has changed.

## Installing

`httptest2` can be installed from CRAN with

```r
install.packages("httptest2")
```

The pre-release version of the package can be pulled from GitHub using the [remotes](https://github.com/r-lib/remotes) package:

```r
# install.packages("remotes")
remotes::install_github("nealrichardson/httptest2")
```

## Getting started

To start using `httptest2` with your package, run `use_httptest2()` in the root of your package directory. This will

* add `httptest2` to "Suggests" in the DESCRIPTION file
* add `library(httptest2)` to `tests/testthat/setup.R`, which `testthat` loads before running tests

Then, you're ready to start using the tools that `httptest2` provides. For an overview of how to get started, see `vignette("httptest2")`, and check out `vignette("faq")` for some common questions. See also the [package reference](https://enpiar.com/httptest2/reference/) for a list of all of the test contexts and expectations provided in the package.

structure(
  list(
    method = "GET",
    url = "http://httpbin.not/basic-auth/user/passwd",
    status_code = 200L,
    headers = structure(
      list(
        Date = "Tue, 30 Nov 2021 20:15:06 GMT",
        `Content-Type` = "application/json",
        `Content-Length` = "47",
        Connection = "keep-alive",
        Server = "gunicorn/19.9.0",
        `Access-Control-Allow-Origin` = "*",
        `Access-Control-Allow-Credentials` = "true"
      ),
      class = "httr2_headers"
    ),
    body = charToRaw(
      "{\n  \"authenticated\": true, \n  \"user\": \"user\"\n}\n"
    )
  ),
  class = "httr2_response"
)

#' `httptest2`: A Test Environment for HTTP Requests
#'
#' The `httptest2` package lets you test R code that wraps an API without
#' requiring access to the remote service. It provides multiple test contexts
#' that mock the network connection in different ways. [with_mock_api()] lets
#' you provide custom fixtures as responses to requests, stored as plain-text
#' files in your test directory. [without_internet()] converts HTTP requests
#' into errors that print the request method, URL, and body payload, if
#' provided, allowing you to assert that a function call would make a
#' correctly-formed HTTP request or assert that a function does not make a
#' request (because if it did, it would raise an error in this context).
#'
#' The package also includes [capture_requests()], a context that collects the
#' responses from requests you make and stores them as mock files. This enables
#' you to perform a series of requests against a live server once and then build
#' your test suite using those mocks, running your tests in `with_mock_api()`.
#' For a context that combines both recording and mocking, use
#' [with_mock_dir()]: it records the first time it is run, and subsequent
#' runs use the mocks for offline testing.
#'
#' `httptest2` also includes expectations to assert that HTTP requests
#' were---or were not---made. [expect_GET()], [expect_PUT()], [expect_PATCH()],
#' [expect_POST()], and [expect_DELETE()] assert that the specified HTTP request
#' is made within one of the test contexts. They catch the error or message
#' raised by the mocked HTTP service and check that the request URL and optional
#' body match the expectation. [expect_no_request()] is the inverse of those: it
#' asserts that no error or message from a mocked HTTP service is raised.
#' [expect_request_header()] asserts that an HTTP request, mocked or not,
#' contains a request header.
#'
#' For an overview of testing with `httptest2`, see `vignette("httptest2")`.
#'
#' When recording requests, by default `httptest2` looks for and redacts the
#' standard ways that auth credentials are passed in requests. This prevents you
#' from accidentally publishing your personal tokens. The redacting behavior is
#' fully customizable, either by providing a `function (response) {...}` to
#' `set_redactor()`, or by placing a function in your package's
#' `inst/httptest2/redact.R` that will be used automatically any time you record
#' requests with your package loaded. See `vignette("redacting")` for details.
#'
#' `httptest2` also enables you to write package vignettes and other R Markdown
#' documents that communicate with a remote API. By adding as little as
#' [start_vignette()] to the beginning of your vignette, you can safely record
#' API responses from a live session, using your secret credentials. These API
#' responses are scrubbed of sensitive personal information and stored in a
#' subfolder in your `vignettes` directory. Subsequent vignette builds,
#' including on continuous-integration services, CRAN, and your package users'
#' computers, use these recorded responses, allowing the document to regenerate
#' without a network connection or API credentials. To record fresh API
#' responses, delete the subfolder of cached responses and re-run. See
#' `vignette("vignettes")` for more discussion and links to examples.
#'
#' @name httptest2
#' @docType package
#' @importFrom httr2 req_perform request
#' @keywords internal
"_PACKAGE"

structure(
  list(
    method = "GET",
    url = "http://httpbin.not/cookies",
    status_code = 200L,
    headers = structure(
      list(
        Date = "Sun, 05 Dec 2021 20:01:43 GMT",
        `Content-Type` = "application/json",
        `Content-Length` = "44",
        Connection = "keep-alive",
        Server = "gunicorn/19.9.0",
        `Access-Control-Allow-Origin` = "*",
        `Access-Control-Allow-Credentials` = "true",
        # Set-Cookie is artificially added here because it is lost in a 302 redirect
        `Set-Cookie` = "token=12345; Domain=example.com; Max-Age=31536000; Path=/"
      ),
      class = "httr2_headers"
    ),
    body = charToRaw("{\n  \"cookies\": {\n    \"token\": \"12345\"\n  }\n}\n")
  ),
  class = "httr2_response"
)

#' Set an alternate directory for mock API fixtures
#'
#' By default, `with_mock_api()` will look for and `capture_requests()` will
#' write mocks to your package's `tests/testthat` directory, or else the current
#' working directory if that path does not exist. If you want to look in or
#' write to other places, call `.mockPaths()` to add directories to the search
#' path.
#'
#' It works like [base::.libPaths()]: any directories you specify will be added
#' to the list and searched first. The default directory will be searched last.
#' Only unique values are kept: if you provide a path that is already found in
#' `.mockPaths()`, the result effectively moves that path to the first position.
#'
#' For finer-grained control, or to completely override the default behavior
#' of searching in the current working directory, you can set the option
#' "httptest2.mock.paths" directly.
#' @param new Either a character vector of path(s) to add, or `NULL` to reset
#' to the default.
#' @return If `new` is omitted, the function returns the current search paths, a
#' a character vector. If `new` is provided, the updated value will be returned
#' invisibly.
#' @examples
#' .mockPaths()
#' .mockPaths("/var/somewhere/else")
#' .mockPaths()
#' .mockPaths(NULL)
#' .mockPaths()
#' @rdname mockPaths
#' @export
.mockPaths <- function(new) {
  current <- getOption("httptest2.mock.paths", default = default_mock_path())
  if (missing(new)) {
    # We're calling the function to get the list of paths
    return(current)
  } else if (is.null(new)) {
    # We're calling the function to reset to the default
    options(httptest2.mock.paths = new)
    invisible(current)
  } else {
    # We're adding one or more paths
    current <- unique(c(new, current))
    options(httptest2.mock.paths = current)
    invisible(current)
  }
}

default_mock_path <- function() {
  if (dir.exists("tests/testthat")) {
    "tests/testthat"
  } else {
    "."
  }
}

with_mock_path <- function(path, expr, replace = FALSE) {
  oldmp <- .mockPaths()
  if (replace) {
    options(httptest2.mock.paths = path)
  } else {
    # Append
    .mockPaths(path)
  }
  on.exit(options(httptest2.mock.paths = oldmp))
  eval.parent(expr)
}

test_that("with_mock_dir works when the directory with mock files exists", {
  temporary_dir <- tempfile()
  # The directory exists, so we'll read mocks from it
  dir.create(temporary_dir)
  # Put mocks in it
  file.copy(testthat::test_path("httpbin.not"), temporary_dir, recursive = TRUE)

  current_mock_paths <- .mockPaths()
  with_mock_dir(temporary_dir, {
    # There's nothing on the mock paths other than our mock dir
    expect_identical(.mockPaths(), temporary_dir)
    expect_no_request(resp <- request("http://httpbin.not/status/204") %>% req_perform())
    # We've read from the mock
    expect_equal(resp_headers(resp)$date, "Sat, 24 Feb 2018 00:22:11 GMT")
  })
  # Original mock paths are reset
  expect_true(all.equal(current_mock_paths, .mockPaths()))
})

test_that("with_mock_dir creates mock files directory", {
  temporary_dir <- tempfile()
  current_mock_paths <- .mockPaths()
  with_mock_dir(temporary_dir, {
    # Since the directory doesn't exist, we will create it and capture
    expect_identical(.mockPaths(), temporary_dir)
    expect_false(dir.exists(temporary_dir))

    resp <- request(httpbin$url("/status/204")) %>% req_perform()
    # It recorded the request in temporary_dir
    expect_true(dir.exists(temporary_dir))
  })
  expect_true(all.equal(current_mock_paths, .mockPaths()))
})

test_that("with_mock_dir prefers to be in tests/testthat", {
  d <- tempfile()
  dir.create(file.path(d, "tests", "testthat"), recursive = TRUE)
  old <- setwd(d)
  on.exit(setwd(old))

  with_mock_dir("asdf", {
    expect_identical(.mockPaths(), file.path("tests", "testthat", "asdf"))
  })
  with_mock_dir("/asdf", {
    expect_identical(.mockPaths(), "/asdf")
  })
})

test_that(".mockPaths works more or less like .libPaths", {
  expect_identical(.mockPaths(), ".")
  .mockPaths("something else")
  expect_identical(.mockPaths(), c("something else", "."))
  # Unique paths
  .mockPaths(".")
  expect_identical(.mockPaths(), c(".", "something else"))
  .mockPaths(NULL)
  expect_identical(.mockPaths(), ".")
})

test_that(".mockPaths default path prefers to be in tests/testthat", {
  d <- tempfile()
  dir.create(file.path(d, "tests", "testthat"), recursive = TRUE)
  old <- setwd(d)
  on.exit(setwd(old))

  expect_identical(.mockPaths(), "tests/testthat")
})

with_mock_api({
  test_that("GET with no query, default mock path", {
    b <- request("https://test.api/object1/") %>% req_perform()
    expect_identical(resp_body_json(b), list(object = TRUE))
  })
  test_that("GET with query, default mock path", {
    obj <- request("https://test.api/object1/") %>%
      req_url_query(a = 1) %>%
      req_perform()
    expect_equal(
      resp_body_json(obj),
      list(query = list(a = 1), mocked = "yes"),
      list_as_map = TRUE
    )
  })
  test_that("There is no api/object2/ mock", {
    expect_GET(request("https://test.api/object2/") %>% req_perform())
  })

  .mockPaths("alt")
  test_that("GET with query, different mock path", {
    obj <- request("https://test.api/object1/") %>%
      req_url_query(a = 1) %>%
      req_perform()
    expect_equal(
      resp_body_json(obj),
      list(query = list(a = 1), mocked = "twice"),
      list_as_map = TRUE
    )
  })
  test_that("Now there is an api/object2/ mock", {
    obj <- request("https://test.api/object2/") %>% req_perform()
    expect_identical(resp_body_json(obj), list(object2 = TRUE))
  })
  test_that("If the primary mock dir doesn't have a mock, it passes to next", {
    b <- request("https://test.api/object1/") %>% req_perform()
    expect_identical(resp_body_json(b), list(object = TRUE))
  })
  test_that("Failure to find a mock in any dir", {
    expect_GET(request("https://test.api/NOTAFILE/") %>% req_perform())
  })

  .mockPaths(NULL)
  test_that("NULL mockPaths resets to default", {
    obj <- request("https://test.api/object1/") %>%
      req_url_query(a = 1) %>%
      req_perform()
    expect_equal(
      resp_body_json(obj),
      list(query = list(a = 1), mocked = "yes"),
      list_as_map = TRUE
    )
    expect_GET(request("https://test.api/object2/") %>% req_perform())
  })
})

structure(
  list(
    url = "http://httpbin.not/response-headers?Location=http%3A%2F%2Fhttpbin.not%2Fstatus%2F201",
    status_code = 200L,
    headers = structure(
      list(
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Sat, 24 Feb 2018 00:22:10 GMT",
        `content-type` = "application/json",
        location = "http://httpbin.not/status/201",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0",
        `content-length` = "89",
        via = "1.1 vegur"
      ),
      .Names = c(
        "connection",
        "server",
        "date",
        "content-type",
        "location",
        "access-control-allow-origin",
        "access-control-allow-credentials",
        "x-powered-by",
        "x-processed-time",
        "content-length",
        "via"
      ),
      class = c("insensitive", "list")
    ),
    all_headers = list(structure(
      list(
        status = 200L,
        version = "HTTP/1.1",
        headers = structure(
          list(
            connection = "keep-alive",
            server = "meinheld/0.6.1",
            date = "Sat, 24 Feb 2018 00:22:10 GMT",
            `content-type` = "application/json",
            location = "http://httpbin.not/status/201",
            `access-control-allow-origin` = "*",
            `access-control-allow-credentials` = "true",
            `x-powered-by` = "Flask",
            `x-processed-time` = "0",
            `content-length` = "89",
            via = "1.1 vegur"
          ),
          .Names = c(
            "connection",
            "server",
            "date",
            "content-type",
            "location",
            "access-control-allow-origin",
            "access-control-allow-credentials",
            "x-powered-by",
            "x-processed-time",
            "content-length",
            "via"
          ),
          class = c("insensitive", "list")
        )
      ),
      .Names = c("status", "version", "headers")
    )),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      .Names = c(
        "domain",
        "flag",
        "path",
        "secure",
        "expiration",
        "name",
        "value"
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(
      "{\n  \"Content-Type\": \"application/json\", \n  \"Location\": \"http://httpbin.not/status/201\"\n}\n"
    ),
    date = structure(1519431730, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = structure(
      c(
        0,
        3.1e-05,
        3.3e-05,
        7.6e-05,
        0.091662,
        0.091723
      ),
      .Names = c(
        "redirect",
        "namelookup",
        "connect",
        "pretransfer",
        "starttransfer",
        "total"
      )
    )
  ),
  .Names = c(
    "url",
    "status_code",
    "headers",
    "all_headers",
    "cookies",
    "content",
    "date",
    "times"
  ),
  class = "response"
)

#' Test that an HTTP request is made with a header
#'
#' This expectation checks that HTTP headers (and potentially header values)
#' are present in a request. It works both in the mock HTTP contexts and on
#' "live" HTTP requests.
#'
#' @param expr Code to evaluate
#' @param ... Named headers to match. Values should either be a string (length-1
#' character), which will be passed to [testthat::expect_match()], or `NULL` to
#' assert that the named header is not present in the request. To assert that a
#' header is merely present in the request, without asserting anything about its
#' contents, provide an empty string (`""`). Header names are always
#' case-insensitive; header values will be matched using the following
#' parameters:
#' @inheritParams base::grepl
#' @return The value of `expr` if there are no expectation failures
#' @importFrom testthat expect_null expect_match
#' @importFrom rlang is_string
#' @examplesIf FALSE
#'
#' library(httr2)
#' expect_request_header(
#'   request("http://httpbin.org") %>%
#'     req_headers(Accept = "image/png") %>%
#'     req_perform(),
#'   accept = "image/png",
#'   `x-fake-header` = NULL
#' )
#' expect_request_header(
#'   request("http://httpbin.org") %>%
#'     req_headers(Accept = "image/png") %>%
#'     req_perform(),
#'   accept = ""
#' )
#' @export
expect_request_header <- function(expr,
                                  ...,
                                  fixed = FALSE,
                                  ignore.case = FALSE,
                                  perl = FALSE,
                                  useBytes = FALSE) {
  expected <- list(...)
  if (length(expected) == 0) {
    stop("No headers provided")
  }
  if (is.null(names(expected)) || !all(nzchar(names(expected)))) {
    stop("Header values must be named")
  }
  names(expected) <- tolower(names(expected))

  # Slightly different behavior based on whether we already are mocking
  current_mocker <- getOption("httr2_mock")
  header_mocker <- function(req) {
    actual <- req$headers
    names(actual) <- tolower(names(actual))
    for (n in names(expected)) {
      if (is.null(expected[[n]])) {
        expect_null(
          actual[[n]],
          label = sprintf('Header "%s"', n)
        )
      } else if (is_string(expected[[n]])) {
        expect_match(
          actual[[n]],
          expected[[n]],
          fixed = fixed,
          ignore.case = ignore.case,
          perl = perl,
          useBytes = useBytes,
          label = sprintf('Header "%s"', n)
        )
      } else {
        stop("Expected headers must be strings (length 1)", call. = FALSE)
      }
    }

    if (is.null(current_mocker)) {
      # If there is no other mocking, req_perform() will proceed normally
      NULL
    } else {
      current_mocker(req)
    }
  }

  with_mock(header_mocker, expr)
}

with_mock <- function(mock, code) {
  if (httr2_1.0.0) {
    utils::getFromNamespace("with_mocked_responses", "httr2")(mock, code)
  } else {
    httr2::with_mock(mock, code)
  }
}

with_mock_api({
  test_that("expect_request_header with mock API", {
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg") %>%
        req_perform(),
      accept = "image/jpeg"
    )
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg", `X-Stuff` = "more") %>%
        req_perform(),
      accept = ""
    )
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg", `X-Stuff` = "more") %>%
        req_perform(),
      Accept = "image/jpeg",
      `X-stuff` = "mo",
      `x-not-present` = NULL
    )
    expect_failure(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform(),
        accept = "image/jpeg"
      ),
      'Header "accept" does not match "image/jpeg"'
    )
    expect_failure(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform(),
        accept = NULL
      ),
      'Header "accept" is not NULL'
    )
  })

  test_that("Args passed to expect_match", {
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg") %>%
        req_perform(),
      accept = "image.*"
    )
    expect_failure(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/jpeg") %>%
          req_perform(),
        accept = "image.*",
        fixed = TRUE
      )
    )
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg") %>%
        req_perform(),
      accept = "IMAGE",
      ignore.case = TRUE
    )
    expect_failure(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/jpeg") %>%
          req_perform(),
        accept = "IMAGE"
      )
    )
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg") %>%
        req_perform(),
      accept = "image.*",
      perl = TRUE
    )
    expect_request_header(
      request("https://test.api/object1/") %>%
        req_headers(Accept = "image/jpeg") %>%
        req_perform(),
      accept = "image",
      useBytes = TRUE
    )
  })

  test_that("Input validation", {
    expect_error(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform()
      ),
      "No headers provided"
    )
    expect_error(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform(),
        "image/png"
      ),
      "Header values must be named"
    )
    expect_error(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform(),
        accept = "image/png",
        "another"
      ),
      "Header values must be named"
    )
    expect_error(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform(),
        accept = c("image", "/png")
      ),
      "Expected headers must be strings (length 1)",
      fixed = TRUE
    )
    expect_error(
      expect_request_header(
        request("https://test.api/object1/") %>%
          req_headers(Accept = "image/png") %>%
          req_perform(),
        accept = c(12, 34)
      ),
      "Expected headers must be strings (length 1)",
      fixed = TRUE
    )
  })
})

without_internet({
  test_that("expect_request_header without_internet", {
    expect_GET(
      expect_request_header(
        request("http://httpbin.not/") %>%
          req_headers(Accept = "image/jpeg") %>%
          req_perform(),
        Accept = "image/jpeg"
      )
    )
    # Because the header check fails before the stop_request() mock
    # is called, there is no GET to assert
    expect_no_request(
      expect_failure(
        expect_request_header(
          request("http://httpbin.not/") %>%
            req_headers(Accept = "image/png") %>%
            req_perform(),
          Accept = "image/jpeg"
        )
      )
    )
  })
})

test_that("expect_request_header works with actual network too", {
  expect_request_header(
    request(httpbin$url("/")) %>%
      req_headers(Accept = "image/jpeg") %>%
      req_perform(),
    Accept = "image/jpeg"
  )
  expect_failure(
    expect_request_header(
      request(httpbin$url("/")) %>%
        req_headers(Accept = "image/png") %>%
        req_perform(),
      Accept = "image/jpeg"
    )
  )
})

structure(
  list(
    url = "http://httpbin.not/status/204",
    status_code = 204L,
    headers = structure(
      list(
        `content-length` = "0",
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Sat, 24 Feb 2018 00:22:11 GMT",
        `content-type` = "text/html; charset=utf-8",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0",
        via = "1.1 vegur"
      ),
      .Names = c(
        "content-length",
        "connection",
        "server",
        "date",
        "content-type",
        "access-control-allow-origin",
        "access-control-allow-credentials",
        "x-powered-by",
        "x-processed-time",
        "via"
      ),
      class = c("insensitive", "list")
    ),
    all_headers = list(
      structure(
        list(
          status = 204L,
          version = "HTTP/1.1",
          headers = structure(
            list(
              `content-length` = "0",
              connection = "keep-alive",
              server = "meinheld/0.6.1",
              date = "Sat, 24 Feb 2018 00:22:11 GMT",
              `content-type` = "text/html; charset=utf-8",
              `access-control-allow-origin` = "*",
              `access-control-allow-credentials` = "true",
              `x-powered-by` = "Flask",
              `x-processed-time` = "0",
              via = "1.1 vegur"
            ),
            .Names = c(
              "content-length",
              "connection",
              "server",
              "date",
              "content-type",
              "access-control-allow-origin",
              "access-control-allow-credentials",
              "x-powered-by",
              "x-processed-time",
              "via"
            ),
            class = c("insensitive", "list")
          )
        ),
        .Names = c(
          "status",
          "version",
          "headers"
        )
      )
    ),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      .Names = c(
        "domain",
        "flag",
        "path",
        "secure",
        "expiration",
        "name",
        "value"
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(""),
    date = structure(
      1519431731,
      class = c(
        "POSIXct",
        "POSIXt"
      ),
      tzone = "GMT"
    ),
    times = structure(
      c(
        0,
        4.5e-05,
        4.8e-05,
        0.000105,
        0.100235,
        0.100262
      ),
      .Names = c(
        "redirect",
        "namelookup",
        "connect",
        "pretransfer",
        "starttransfer",
        "total"
      )
    )
  ),
  .Names = c(
    "url",
    "status_code",
    "headers",
    "all_headers",
    "cookies",
    "content",
    "date",
    "times"
  ),
  class = "response"
)

#' Remove sensitive content from HTTP responses
#'
#' When recording requests for use as test fixtures, you don't want to include
#' secrets like authentication tokens and personal ids. These functions provide
#' a means for redacting this kind of content, or anything you want, from
#' responses that [capture_requests()] saves.
#'
#' `redact_cookies()` removes cookies from `httr2_response` objects
#' and is the default redactor in `capture_requests()`.
#' `redact_headers()` lets you target selected request and response headers for
#' redaction.
#' `within_body_text()` lets you manipulate the text of the response body
#' and manages the parsing of the raw (binary) data in the `httr_response` object.
#'
#' Note that if you set a redacting function, it will also be applied to requests when loading mocks. This allows you to sanitize and/or shorten URLs in your mock files.
#' @param response An `httr2_response` or `httr2_request` object to sanitize.
#' @param headers For `redact_headers()`, a character vector of header names to
#' sanitize.
#' @param FUN For `within_body_text()`, a function that takes as its argument a
#' character vector and returns a modified version of that. This function will
#' be applied to the text of the response's body.
#' @return All redacting functions return a well-formed `httr2_response`
#' or `httr2_request` object.
#' @name redact
#' @aliases redact_cookies redact_headers within_body_text
#' @seealso `vignette("redacting", package = "httptest2")` for a detailed discussion of what these functions do and how to customize them. [gsub_response()] is another redactor.
#' @export
redact_cookies <- function(response) {
  redact_headers(response, "Set-Cookie")
}

#' @rdname redact
#' @export
redact_headers <- function(response, headers = c()) {
  header_apply(response, headers, function(x) "REDACTED")
}

header_apply <- function(response, headers, FUN, ...) {
  # Apply some function over a set of named headers
  todo <- tolower(names(response$headers)) %in% tolower(headers)
  response$headers[todo] <- lapply(response$headers[todo], FUN, ...)
  response
}

#' @rdname redact
#' @export
within_body_text <- function(response, FUN) {
  # This could be applied to a httr2_request object;
  # only do work on a httr2_response
  if (inherits(response, "httr2_response") && length(response$body)) {
    old <- resp_body_string(response)
    new <- FUN(old)
    response$body <- charToRaw(new)
  }
  return(response)
}

#' Find and replace within a response object
#'
#' This function passes its arguments to [base::gsub()] in order to find and
#' replace string patterns (regular expressions) within
#' the URL and the response body of `httr2_response` objects.
#'
#' Note that, unlike `gsub()`, the first argument of the function is `response`,
#' not `pattern`, while the equivalent argument in `gsub()`, "`x`", is placed
#' third. This difference is to maintain consistency with the other redactor
#' functions in `httptest2`, which all take `response` as the first argument.
#'
#' This function also can be applied to an `http2_request` object to replace
#' patterns inside the request URL.
#' @param response An `httr2_response` or `http2_request` object to sanitize.
#' @param pattern From [base::gsub()]: "character string containing a regular
#' expression (or character string for `fixed = TRUE`) to be matched in the
#' given character vector." Passed to `gsub()`. See the docs for `gsub()` for
#' further details.
#' @param replacement A replacement for the matched pattern, possibly including
#' regular expression backreferences. Passed to `gsub()`. See the docs for
#' `gsub()` for further details.
#' @param ... Additional logical arguments passed to `gsub()`: `ignore.case`,
#' `perl`, `fixed`, and `useBytes` are the possible options.
#' @return An `httr2_response` object, same as was passed in, with the
#' pattern replaced in the URLs and bodies.
#' @export
gsub_response <- function(response, pattern, replacement, ...) {
  replacer <- function(x) gsub(pattern, replacement, x, ...)
  response$url <- replacer(response$url)
  response <- header_apply(response, "location", replacer)
  response <- within_body_text(response, replacer)
  return(response)
}

#' Wrap a redacting expression as a proper function
#'
#' Redactors take a `httr2_response` or `httr2_request` as their first argument, and some take additional
#' arguments: `redact_headers()`, for example, requires that you specify
#' `headers`. This function allows you to take a simplified expression via a
#' formula, similar to what `purrr` does, so that you can provide the function
#' to `capture_requests()`.
#'
#' For example, `as.redactor(~ redact_headers(., "X-Custom-Header"))` is equivalent
#' to `function (response) redact_headers(response, "X-Custom-Header")`. This
#' allows you to do
#' `set_redactor(~ redact_headers(., "X-Custom-Header"))`.
#' @param fmla Partial expression (a `formula`) to turn into a function
#' @return A `function`.
#' @rdname as-redactor
#' @importFrom stats terms
#' @keywords internal
#' @seealso [capture_requests()]
as.redactor <- function(fmla) {
  env <- parent.frame()
  expr <- attr(terms(fmla), "variables")[[2]]
  # cf. magrittr:::wrap_function: wrap that in a function (.) ...
  return(eval(call("function", as.pairlist(alist(. = )), expr), env, env))
}

chain_redactors <- function(funs) {
  # Given a list of functions, return a function that execs them in sequence
  return(function(response) {
    for (f in funs) {
      if (inherits(f, "formula")) {
        f <- as.redactor(f)
      }
      response <- f(response)
    }
    return(response)
  })
}

d <- tempfile()

with_mock_api({
  # Auth headers aren't recorded
  capture_while_mocking(simplify = FALSE, path = d, {
    a <- request("https://test.api/") %>%
      req_headers(`Authorization` = "Bearer token") %>%
      req_perform()
  })
  test_that("The mock file does not have the request headers", {
    # In httr2, response objects do not include the request,
    # so by construction there won't be request headers
    expect_false(any(grepl("Bearer token", readLines(file.path(d, "test.api.R")))))
  })
  test_that("And the redacted .R mock can be loaded", {
    with_mock_path(d, {
      b <- request("https://test.api/") %>%
        req_headers(`Authorization` = "Bearer token") %>%
        req_perform()
    })
    expect_equal(resp_body_json(b), resp_body_json(a))
  })

  # redact_cookies from response
  c2_req <- request("http://httpbin.not/cookies/set") %>%
    req_url_query(token = 12345)
  test_that("redact_cookies: the response has the set-cookie in the response", {
    capture_while_mocking(simplify = FALSE, path = d, {
      # Slight hack: httpbin does a 302 redirect, and the Set-Cookie appears
      # in the 302, but the final 200 response doesn't have it, so I've
      # modified the mock to have the Set-Cookie header, just so we can test
      # redacting it.
      # This could also be done by setting `req_options(followlocation = FALSE)`
      c2 <- req_perform(c2_req)
      expect_equal(
        resp_header(c2, "set-cookie"),
        "token=12345; Domain=example.com; Max-Age=31536000; Path=/"
      )
    })
  })
  test_that("redact_cookies removes set-cookies from response in the mock file", {
    expect_length(
      grep(
        "REDACTED",
        readLines(file.path(d, "httpbin.not", "cookies", "set-5b2631.R"))
      ),
      1
    )
  })
  test_that("And when loading that .R mock, the redacted value doesn't appear", {
    with_mock_path(d, {
      expect_identical(
        req_perform(c2_req) %>% resp_header("set-cookie"),
        "REDACTED"
      )
    })
  })

  # HTTP auth credentials aren't recorded
  auth_req <- request("http://httpbin.not/basic-auth/user/passwd") %>%
    # Slight hack: the mock was recorded with user:passwd to get a 200 response
    # but then the grepl would fail because the password is in the URL
    # (a feature of httpbin)
    req_auth_basic("user", "SeCrEtPaSsWoRd!") %>%
    req_error(is_error = ~FALSE)
  capture_while_mocking(simplify = FALSE, path = d, {
    pwauth <- req_perform(auth_req)
  })
  test_that("there is no password in the mock", {
    expect_false(any(grepl(
      "SeCrEtPaSsWoRd!",
      readLines(file.path(d, "httpbin.not", "basic-auth", "user", "passwd.R"))
    )))
  })
  test_that("And the redacted .R mock can be loaded", {
    with_mock_path(d, {
      pwauthb <- req_perform(auth_req)
    })
    expect_equal(resp_body_json(pwauthb), resp_body_json(pwauth))
  })

  # Custom redacting function
  my_redactor <- function(response) {
    # Proof that you can alter other parts of the response/mock when recording
    # Slight finesse: because requests get preprocessed with the redactor too,
    # it's tricky when we're mocking and recording, so in this test we're not
    # to change the URL when determining the mock file to load
    if (!grepl("get_current_redactor()(req)", unlist(tail(sys.calls(), 1)), fixed = TRUE)) {
      response$url <- "http://example.com/fakeurl"
    }
    # Proof that you can alter the response body
    cleaner <- function(x) gsub("loaded", "changed", x)
    response <- within_body_text(response, cleaner)
    return(response)
  }
  with_redactor(
    my_redactor,
    capture_while_mocking(simplify = FALSE, path = d, {
      r <- request("http://example.com/get") %>% req_perform()
    })
  )
  test_that("The real request is not affected by the redactor", {
    expect_identical(r$url, "http://example.com/get")
    expect_identical(resp_body_json(r), list(loaded = TRUE))
  })
  test_that("But the mock file gets written to the modified path with altered content", {
    # Use replace=TRUE to make sure that "." isn't in the search path.
    # We're checking that the original request doesn't have a mock,
    # but of course we made it from a mock in the working directory
    with_mock_path(d, replace = TRUE, {
      expect_GET(
        request("http://example.com/get") %>% req_perform(),
        "http://example.com/get"
      )
      expect_error(alt <- request("http://example.com/fakeurl") %>% req_perform(), NA)
      expect_identical(resp_body_json(alt), list(changed = TRUE))
    })
  })

  # New in httptest2: redactor is used as request preprocessor/URL shortener
  test_that("Redactors are applied when making requests to alter the mock file path we're reading", {
    with_redactor(
      function(resp) gsub_response(resp, "long/url.*$", "get"),
      r <- request("http://example.com/long/url/with/lots/of/segments") %>% req_perform()
    )
    # The URL of the mock response in this case is actually the full request URL
    # because it is a JSON mock so the httr2_response object is generated
    # based on the request
    expect_identical(r$url, "http://example.com/long/url/with/lots/of/segments")
    # But this is the response body of the mock corresponding to example.com/get
    expect_identical(resp_body_json(r), list(loaded = TRUE))
  })

  a <- request("https://test.api/") %>%
    req_headers(`Authorization` = "Bearer token") %>%
    req_perform()
  test_that("gsub_response", {
    asub <- gsub_response(a, "api", "OTHER")
    expect_identical(asub$url, "https://test.OTHER/")
    expect_identical(resp_body_json(asub), list(value = "https://test.OTHER/object1/"))
  })
  test_that("as.redactor", {
    a2 <- prepare_redactor(~ gsub_response(., "api", "OTHER"))(a)
    expect_identical(resp_body_json(a2), list(value = "https://test.OTHER/object1/"))
  })

  loc <- request("http://httpbin.not/response-headers") %>%
    req_url_query(Location = "http://httpbin.not/status/201") %>%
    req_perform()
  loc_sub <- gsub_response(
    loc, "http://httpbin.not/status/201",
    "http://httpbin.not/status/404"
  )
  test_that("gsub_response touches Location header", {
    expect_identical(
      resp_header(loc_sub, "location"),
      "http://httpbin.not/status/404"
    )
    expect_identical(
      resp_body_json(loc_sub)$Location,
      "http://httpbin.not/status/404"
    )
  })
  test_that("gsub_response handles URL encoding", {
    skip("TODO: handle URL escaping")
    expect_identical(
      loc_sub$url,
      "http://httpbin.not/response-headers?Location=http%3A%2F%2Fhttpbin.not%2Fstatus%2F404"
    )
  })
})

test_that("gsub_response handles empty response bodies (#20)", {
  with_redactor(
    function(resp) gsub_response(resp, "status", "code"),
    capture_requests({
      # Prior to the fix, the redactor errored here on retrieving an empty body
      r <- request(httpbin$url("/status/204")) %>% req_perform()
    })
  )
  # Nothing here
  expect_length(r$body, 0)
})

test_that("chain_redactors", {
  f1 <- function(x) x * 4
  f2 <- ~ sum(c(., 3))
  f12 <- chain_redactors(list(f1, f2))
  f21 <- chain_redactors(list(f2, f1))
  expect_equal(f12(5), 23)
  expect_equal(f21(5), 32)
})

reset_redactors()

structure(
  list(
    url = "http://httpbin.not/cookies",
    status_code = 200L,
    headers = structure(
      list(
        connection = "keep-alive",
        server = "meinheld/0.6.1",
        date = "Thu, 14 Sep 2017 04:34:50 GMT",
        `content-type` = "application/json",
        `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true",
        `x-powered-by` = "Flask",
        `x-processed-time` = "0.000942945480347",
        `content-length` = "44",
        via = "1.1 vegur"
      ),
      class = c(
        "insensitive",
        "list"
      )
    ),
    all_headers = list(list(
      status = 200L,
      version = "HTTP/1.1",
      headers = structure(
        list(
          connection = "keep-alive",
          server = "meinheld/0.6.1",
          date = "Thu, 14 Sep 2017 04:34:50 GMT",
          `content-type` = "application/json",
          `access-control-allow-origin` = "*",
          `access-control-allow-credentials` = "true",
          `x-powered-by` = "Flask",
          `x-processed-time` = "0.000942945480347",
          `content-length` = "44",
          via = "1.1 vegur"
        ),
        class = c(
          "insensitive",
          "list"
        )
      )
    )),
    cookies = structure(
      list(
        domain = logical(0),
        flag = logical(0),
        path = logical(0),
        secure = logical(0),
        expiration = structure(
          numeric(0),
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = logical(0),
        value = logical(0)
      ),
      row.names = integer(0),
      class = "data.frame"
    ),
    content = charToRaw(
      "{\n  \"cookies\": {\n    \"token\": \"12345\"\n  }\n}\n"
    ),
    date = structure(1505363690, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = c(
      redirect = 0,
      namelookup = 0.123097,
      connect = 0.34047,
      pretransfer = 0.340592,
      starttransfer = 0.546105,
      total = 0.546146
    ),
    request = structure(
      list(
        method = "GET",
        url = "http://httpbin.not/cookies",
        headers = c(
          Accept = "application/json, text/xml, application/xml, */*"
        ),
        fields = NULL,
        options = list(
          useragent = "libcurl/7.54.0 r-curl/2.8.1 httr/1.3.1",
          cookie = "token=12345",
          httpget = TRUE
        ),
        auth_token = NULL,
        output = structure(
          list(),
          class = c(
            "write_memory",
            "write_function"
          )
        )
      ),
      class = "request"
    )
  ),
  class = "response"
)

structure(
  list(
    url = "http://httpbin.not/html",
    status_code = 200L,
    headers = structure(list(
      connection = "keep-alive",
      server = "meinheld/0.6.1",
      date = "Sat, 27 May 2017 20:08:33 GMT",
      `content-type` = "text/html; charset=utf-8",
      `content-length` = "3741",
      `access-control-allow-origin` = "*",
      `access-control-allow-credentials` = "true",
      `x-powered-by` = "Flask",
      `x-processed-time` = "0.000486850738525",
      via = "1.1 vegur"
    ), class = c("insensitive", "list")),
    all_headers = list(
      structure(list(status = 200L, version = "HTTP/1.1", headers = structure(list(
        connection = "keep-alive", server = "meinheld/0.6.1",
        date = "Sat, 27 May 2017 20:08:33 GMT", `content-type` = "text/html; charset=utf-8",
        `content-length` = "3741", `access-control-allow-origin` = "*",
        `access-control-allow-credentials` = "true", `x-powered-by` = "Flask",
        `x-processed-time` = "0.000486850738525", via = "1.1 vegur"
      ), .Names = c(
        "connection",
        "server", "date", "content-type", "content-length", "access-control-allow-origin",
        "access-control-allow-credentials", "x-powered-by", "x-processed-time",
        "via"
      ), class = c("insensitive", "list"))), .Names = c(
        "status",
        "version", "headers"
      ))
    ),
    content = charToRaw("<!DOCTYPE html>
<html>
  <head>
    </head>
      <body>
        <h1>Herman Melville - Moby-Dick</h1>
            <div>
                <p>Availing himself of the mild, summer-cool weather that now reigned in these latitudes, and in preparation for the peculiarly active pursuits shortly to be anticipated, Perth, the begrimed, blistered old blacksmith, had not removed his portable forge to the hold again, after concluding his contributory work for Ahab's leg, but still retained it on deck, fast lashed to ringbolts by the foremast; being now almost incessantly invoked by the headsmen, and harpooneers, and bowsmen to do some little job for them; altering, or repairing, or new shaping their various weapons and boat furniture. Often he would be surrounded by an eager circle, all waiting to be served; holding boat-spades, pike-heads, harpoons, and lances, and jealously watching his every sooty movement, as he toiled. Nevertheless, this old man's was a patient hammer wielded by a patient arm. No murmur, no impatience, no petulance did come from him. Silent, slow, and solemn; bowing over still further his chronically broken back, he toiled away, as if toil were life itself, and the heavy beating of his hammer the heavy beating of his heart. And so it was.—Most miserable! A peculiar walk in this old man, a certain slight but painful appearing yawing in his gait, had at an early period of the voyage excited the curiosity of the mariners. And to the importunity of their persisted questionings he had finally given in; and so it came to pass that every one now knew the shameful story of his wretched fate. Belated, and not innocently, one bitter winter's midnight, on the road running between two country towns, the blacksmith half-stupidly felt the deadly numbness stealing over him, and sought refuge in a leaning, dilapidated barn. The issue was, the loss of the extremities of both feet. Out of this revelation, part by part, at last came out the four acts of the gladness, and the one long, and as yet uncatastrophied fifth act of the grief of his life's drama. He was an old man, who, at the age of nearly sixty, had postponedly encountered that thing in sorrow's technicals called ruin. He had been an artisan of famed excellence, and with plenty to do; owned a house and garden; embraced a youthful, daughter-like, loving wife, and three blithe, ruddy children; every Sunday went to a cheerful-looking church, planted in a grove. But one night, under cover of darkness, and further concealed in a most cunning disguisement, a desperate burglar slid into his happy home, and robbed them all of everything. And darker yet to tell, the blacksmith himself did ignorantly conduct this burglar into his family's heart. It was the Bottle Conjuror! Upon the opening of that fatal cork, forth flew the fiend, and shrivelled up his home. Now, for prudent, most wise, and economic reasons, the blacksmith's shop was in the basement of his dwelling, but with a separate entrance to it; so that always had the young and loving healthy wife listened with no unhappy nervousness, but with vigorous pleasure, to the stout ringing of her young-armed old husband's hammer; whose reverberations, muffled by passing through the floors and walls, came up to her, not unsweetly, in her nursery; and so, to stout Labor's iron lullaby, the blacksmith's infants were rocked to slumber. Oh, woe on woe! Oh, Death, why canst thou not sometimes be timely? Hadst thou taken this old blacksmith to thyself ere his full ruin came upon him, then had the young widow had a delicious grief, and her orphans a truly venerable, legendary sire to dream of in their after years; and all of them a care-killing competency.
                </p>
          </div>
  </body>
</html>"),
    date = structure(1495915713, class = c("POSIXct", "POSIXt"), tzone = "GMT"),
    times = structure(c(0, 0.018544, 0.100954, 0.101057, 0.186309, 0.186663),
      .Names = c(
        "redirect", "namelookup", "connect", "pretransfer",
        "starttransfer", "total"
      )
    ),
    request = structure(
      list(
        method = "GET",
        url = "http://httpbin.not/html",
        headers = structure("application/json, text/xml, application/xml, */*",
          .Names = "Accept"
        ),
        fields = NULL,
        options = structure(list(
          useragent = "libcurl/7.51.0 r-curl/2.3 httr/1.2.1",
          httpget = TRUE
        )),
        auth_token = NULL,
        output = structure(list(), class = c("write_memory", "write_function"))
      ),
      class = "request"
    )
  ),
  class = "response"
)

structure(
  list(
    url = "https://example.com/login/",
    status_code = 204L,
    headers = structure(
      list(
        allow = "GET, HEAD, OPTIONS, POST",
        `content-type` = "application/json;charset=utf-8",
        date = "Thu, 14 Sep 2017 04:27:22 GMT",
        server = "nginx",
        `set-cookie` = "token=12345; Domain=example.com; Max-Age=31536000; Path=/",
        vary = "Cookie, Accept-Encoding",
        connection = "keep-alive"
      ),
      class = c(
        "insensitive",
        "list"
      )
    ),
    all_headers = list(list(
      status = 204L,
      version = "HTTP/1.1",
      headers = structure(
        list(
          allow = "GET, HEAD, OPTIONS, POST",
          `content-type` = "application/json;charset=utf-8",
          date = "Thu, 14 Sep 2017 04:27:22 GMT",
          server = "nginx",
          `set-cookie` = "token=12345; Domain=example.com; Max-Age=31536000; Path=/",
          vary = "Cookie, Accept-Encoding",
          connection = "keep-alive"
        ),
        class = c(
          "insensitive",
          "list"
        )
      )
    )),
    cookies = structure(
      list(
        domain = "example.com",
        flag = TRUE,
        path = "/",
        secure = FALSE,
        expiration = structure(
          1536899241,
          class = c(
            "POSIXct",
            "POSIXt"
          )
        ),
        name = "token",
        value = "12345"
      ),
      row.names = c(
        NA,
        -1L
      ),
      class = "data.frame"
    ),
    content = raw(0),
    date = structure(
      1505363242,
      class = c(
        "POSIXct",
        "POSIXt"
      ),
      tzone = "GMT"
    ),
    times = c(
      redirect = 0,
      namelookup = 0.115726,
      connect = 0.467124,
      pretransfer = 1.405551,
      starttransfer = 2.041081,
      total = 2.041137
    ),
    request = structure(
      list(
        method = "POST",
        url = "https://example.com/login/",
        headers = c(
          Accept = "application/json, text/xml, application/xml, */*",
          `Content-Type` = "application/json",
          `user-agent` = "libcurl/7.54.0 curl/2.8.1 httr/1.3.1"
        ),
        fields = NULL,
        options = list(
          useragent = "libcurl/7.54.0 r-curl/2.8.1 httr/1.3.1",
          post = TRUE,
          postfieldsize = 46L,
          postfields = charToRaw('{"username":"password"}'),
          postredir = 3
        ),
        auth_token = NULL,
        output = structure(
          list(),
          class = c(
            "write_memory",
            "write_function"
          )
        )
      ),
      class = "request"
    )
  ),
  class = "response"
)

#' Set a response redactor
#'
#' A redactor is a function that alters the response content being written
#' out in the [capture_requests()] context, allowing you to remove sensitive
#' values, such as authentication tokens,
#' as well as any other modification or truncation of the response body. By
#' default, the [redact_cookies()] function will be used to purge standard
#' auth methods, but `set_redactor()` allows you to provide a different one.
#'
#' Alternatively, you can put a redacting function in `inst/httptest2/redact.R`
#' in your package, and
#' any time your package is loaded (as in when running tests or building
#' vignettes), the function will be used automatically.
#'
#' @param FUN A function or expression that modifies `httr2_response` objects.
#' Specifically, a valid input is one of:
#' * A function taking a single argument, the `httr2_response`,
#' and returning a valid `httr2_response` object.
#' * A formula as shorthand for an anonymous function with `.` as the
#' "response" argument, as in the `purrr` package. That is, instead of
#' `function (response) redact_headers(response, "X-Custom-Header")`, you can
#' use `~ redact_headers(., "X-Custom-Header")`
#' * A list of redacting functions/formulas, which will be executed
#' in sequence on the response
#' * `NULL`, to override the default `redact_cookies()`.
#' @return Invisibly, the redacting function, validated and perhaps modified.
#' Formulas and function lists are turned into proper functions. `NULL` as input
#' returns the `force()` function.
#' @export
#' @seealso For further details on how to redact responses, see `vignette("redacting", package = "httptest2")`.
#' @examples
#' # Shorten UUIDs in response body/URLs to their first 6 digits:
#' set_redactor(function(resp) gsub_response(resp, "([0-9a-f]{6}[0-9a-f]{26}", "\\1"))
#' # Restore the default
#' set_redactor(redact_cookies)
set_redactor <- function(FUN) {
  FUN <- prepare_redactor(FUN)
  options(
    httptest2.redactor = FUN,
    # Because we're directly setting a redactor, remove any record that
    # a previous redactor was set by reading from packages
    httptest2.redactor.packages = NULL
  )
  invisible(FUN)
}

default_redactor <- function(packages = get_attached_packages()) {
  # Look for package-defined redactors
  func <- redactor_from_packages(packages)
  # Record what packages we considered here
  options(httptest2.redactor.packages = packages)
  return(func)
}

redactor_from_packages <- function(packages) {
  funcs <- find_package_functions(packages, "redact.R")
  if (length(funcs)) {
    out <- prepare_redactor(funcs)
  } else {
    # Default
    out <- redact_cookies
  }
  return(out)
}

find_package_functions <- function(packages, file = "redact.R") {
  # Given package names, find any redactors put in inst/httptest/redact.R
  base_pkgs <- c(
    "base", "compiler", "datasets", "graphics", "grDevices",
    "grid", "methods", "parallel", "splines", "stats", "stats4",
    "tcltk", "tools", "utils"
  )
  packages <- setdiff(packages, base_pkgs)
  funcs <- lapply(packages, get_package_function, file)
  # Make sure we have functions
  funcs <- Filter(is.function, funcs)
  return(funcs)
}

get_package_function <- function(package, file = "redact.R") {
  if ("pkgload" %in% loadedNamespaces()) {
    # Someone may have loaded a package with pkgload::load_all(), so we
    # need this shim function to look up system files
    system.file <- get("shim_system.file", asNamespace("pkgload"))
  }
  func_file <- system.file("httptest2", file, package = package)
  # If file does not exist, it returns ""
  if (!nzchar(func_file)) {
    # Try the httptest dir too
    func_file <- system.file("httptest", file, package = package)
  }
  if (nzchar(func_file)) {
    func <- source(func_file)$value
    if (is.function(func)) {
      verbose_message(paste("Using", file, "from", dQuote(package)))
      return(func)
    }
  }
  return(NULL)
}

#' Fetch the active redacting function
#'
#' Called inside [capture_requests()]. If using the default redactor, it checks
#' each time it is called to see if any new packages have been attached, in case
#' there are package redactors in them.
#' @return A redacting function.
#' @export
#' @keywords internal
get_current_redactor <- function() {
  # First, check where we've cached the current one
  out <- getOption("httptest2.redactor")
  if (is.null(out)) {
    # Set the default
    out <- default_redactor()
    options(httptest2.redactor = out)
  } else {
    # See if default is based on packages and needs refreshing
    pkgs <- getOption("httptest2.redactor.packages")
    if (!is.null(pkgs)) {
      # We're using the result of default_redactor(). Let's see if any
      # new packages have been loaded
      current_packages <- get_attached_packages()
      # Also, always reevaluate the default redactor if pkgload is involved
      if ("pkgload" %in% loadedNamespaces() || !identical(current_packages, pkgs)) {
        # Re-evaluate
        out <- default_redactor(current_packages)
        options(httptest2.redactor = out)
      }
    }
  }
  return(out)
}

prepare_redactor <- function(redactor) {
  if (is.null(redactor)) {
    # Allow, and make it do nothing
    redactor <- force
  } else if (inherits(redactor, "formula")) {
    redactor <- as.redactor(redactor)
  } else if (is.list(redactor)) {
    if (length(redactor) == 1) {
      redactor <- redactor[[1]]
    } else {
      redactor <- chain_redactors(redactor)
    }
  }

  if (!is.function(redactor)) {
    stop("Redactor must be a function or list of functions", call. = FALSE)
  }
  return(redactor)
}

get_attached_packages <- function() {
  gsub("^package\\:", "", grep("^package\\:", search(), value = TRUE))
}

# Some constants

EXT_TO_CONTENT_TYPE <- list(
  "json" = "application/json",
  "xml" = "application/xml",
  "csv" = "text/csv",
  "html" = "text/html",
  "txt" = "text/plain",
  "tsv" = "text/tab-separated-values"
)

CONTENT_TYPE_TO_EXT <- structure(as.list(names(EXT_TO_CONTENT_TYPE)),
  .Names = unlist(EXT_TO_CONTENT_TYPE, use.names = FALSE)
)

#' Convert a request to a mock file path
#'
#' Requests are translated to mock file paths according to several rules that
#' incorporate the request method, URL, query parameters, and body.
#'
#' First, the request protocol, such as "https://", is removed from the URL.
#' Second, if the request URL contains a query string, it will be popped off,
#' hashed by [digest::digest()], and the first six characters appended to the
#' file being read. Third, request bodies are similarly hashed and
#' appended. Finally, if a request method other than GET is used it will be
#' appended to the end of the end of the file name.
#'
#' Mock file paths also have a file extension appended, based on the
#' `Content-Type` of the response, though this function, which is only concerned
#' with the request, does not add the extension. In an
#' HTTP API, a "directory" itself is a resource,
#' so the extension allows distinguishing directories and files in the file
#' system. That is, a mocked `GET http://example.com/api/` may read a
#' "example.com/api.json" file, while
#' `GET http://example.com/api/object1/` reads "example.com/api/object1.json".
#'
#' Other examples:
#' * `GET http://example.com/api/object1/?a=1` may read
#' "example.com/api/object1-b64371.xml".
#' * `POST http://example.com/api/object1/?a=1` may read
#' "example.com/api/object1-b64371-POST.json".
#'
#' Note that if you are trying to guess the mock file paths corresponding to a
#' test for which you intend to create a mock file manually,
#' instead of trying to build the URL, you should run the test
#' with `with_mock_api()` as the error message will contain the mock file path.
#'
#' @param req A `httr2_request` object
#' @return A file path and name, without an extension. The file, or a file with
#' some extension appended, may or may not
#' exist: existence is not a concern of this function.
#' @importFrom digest digest
#' @seealso [with_mock_api()] [capture_requests()]
#' @export
#' @keywords internal
build_mock_url <- function(req) {
  method <- get_request_method(req)
  body <- get_string_request_body(req)

  # Remove protocol
  url <- sub("^.*?://", "", req$url)
  # Handle query params
  parts <- unlist(strsplit(url, "?", fixed = TRUE))
  # Remove trailing slash
  f <- sub("\\/$", "", parts[1])
  # Sanitize the path to be portable for all R platforms
  f <- gsub(":", "-", f)
  if (length(parts) > 1) {
    # There's a query string. Append the digest as a suffix.
    f <- paste0(f, "-", hash(parts[2]))
  }

  # Handle body and append its hash if present
  if (!is.null(body)) {
    f <- paste0(f, "-", hash(body))
  }

  if (method != "GET") {
    # Append method to the file name for non GET requests
    f <- paste0(f, "-", method)
  }
  return(f)
}

get_string_request_body <- function(req) {
  # Returns a string if the request has a body, NULL otherwise
  body_apply <- utils::getFromNamespace("req_body_apply", "httr2")
  req <- body_apply(req)

  b <- request_postfields(req)
  if (is.null(b)) {
    if (length(req$fields)) {
      b <- lapply(req$fields, function(x) {
        if (inherits(x, "form_file")) {
          # hash the file contents
          paste("File:", digest(x$path, serialize = FALSE, file = TRUE))
        } else {
          # assume form_data
          rawToChar(x$value)
        }
      })
      b <- paste(c(
        "Multipart form:",
        paste(names(b), b, sep = " = ")
      ), collapse = "\n  ")
      # add a newline at the end too
      b <- paste0(b, "\n")
    } else if (inherits(req$body$data, "httr2_path")) {
      b <- paste("File:", digest(req$body$data, serialize = FALSE, file = TRUE))
    }
  }
  b
}

request_postfields <- function(req) {
  b <- req[["options"]][["postfields"]]
  if (length(b) > 0) {
    # Check length this way because b may be NULL or length 0 raw vector
    return(rawToChar(b))
  } else {
    return(NULL)
  }
}

get_request_method <- function(req) {
  # At the time that we process the request, some defaults may not have been
  # applied, and the request method may be NULL
  utils::getFromNamespace("req_method_get", "httr2")(req)
}

hash <- function(string, n = 6) substr(digest(string), 1, n)

#' Make all HTTP requests raise an error
#'
#' `without_internet()` simulates the situation when any network request will
#' fail, as in when you are without an internet connection. Any HTTP request
#' through `httr2` will raise an error.
#'
#' `block_requests()` and `stop_mocking()` allow you to turn on/off request
#' blocking for more convenient use in an interactive session.
#'
#' The error message raised has a well-defined shape, made of three
#' elements, separated by space: (1) the request
#' method (e.g. "GET"); (2) the request URL; and
#' (3) the request body, if present. The verb-expectation functions,
#' such as [expect_GET()] and [expect_POST()], look for this shape.
#' @param expr Code to run inside the mock context
#' @return The result of `expr`
#' @examples
#' library(httr2)
#' library(testthat, warn.conflicts = FALSE)
#' without_internet({
#'   expect_error(
#'     request("http://httpbin.org/get") %>% req_perform(),
#'     "GET http://httpbin.org/get"
#'   )
#'   expect_error(
#'     request("http://httpbin.org/put") %>%
#'       req_method("PUT") %>%
#'       req_body_json(list(a = 1)) %>%
#'       req_perform(),
#'     'PUT http://httpbin.org/put {"a":1}',
#'     fixed = TRUE
#'   )
#' })
#' @export
without_internet <- function(expr) with_mock(stop_request, expr)

#' @rdname without_internet
#' @export
block_requests <- function() options(httr2_mock = stop_request)

#' @importFrom rlang abort cnd_header cnd_body cnd_footer
stop_request <- function(req) {
  out <- paste(get_request_method(req), req$url)
  body <- get_string_request_body(req)
  if (!is.null(body)) {
    # Max print option for debugging large payloads
    body <- substr(body, 1, getOption("httptest2.max.print", nchar(body)))
    out <- paste(out, body)
  }

  # Raise an error with a special class attached so we can distinguish it
  abort(out, mockfile = req$mockfile, class = "httptest2_request")
}

#' @export
cnd_header.httptest2_request <- function(cnd, ...) {
  "An unexpected request was made:"
}

#' @export
cnd_body.httptest2_request <- function(cnd, ...) {
  cnd$message
}

#' @export
cnd_footer.httptest2_request <- function(cnd, ...) {
  if (is.null(cnd$mockfile)) {
    character(0)
  } else {
    # Poked in here by mock_request for ease of debugging
    paste0("Expected mock file: ", cnd$mockfile, ".*")
  }
}

#' Wrapper around 'trace' to untrace when finished
#'
#' @param x Name of function to trace. See [base::trace()].
#' @param where where to look for the function to be traced.
#' @param print Logical: print a message every time the traced function is hit?
#' Default is `FALSE`; note that in `trace`, the default is `TRUE`.
#' @param ... Additional arguments to pass to `trace`. At minimum, must include
#' either `tracer` or `exit`.
#' @param expr Code to run inside the context
#' @return The result of `expr`
#' @export
#' @keywords internal
with_trace <- function(x, where = topenv(parent.frame()), print = getOption("httptest2.debug.trace", FALSE), ..., expr) {
  quietly(trace(x, print = print, where = where, ...))
  on.exit(safe_untrace(x, where = where))
  eval.parent(expr)
}

#' @importFrom utils sessionInfo
trace_httr2 <- function(..., print = getOption("httptest2.debug.trace", FALSE)) {
  # Trace it as seen from within httr2
  quietly(trace(..., print = print, where = request))
  # And if httr2 is attached and the function is exported, trace the
  # function as the user sees it
  if ("httr2" %in% names(sessionInfo()$otherPkgs) && ..1 %in% getNamespaceExports("httr2")) {
    try(quietly(trace(..., print = print, where = sys.frame())))
  }
}

quietly <- function(expr) {
  env <- parent.frame()
  if (getOption("httptest2.debug.trace", FALSE)) {
    eval(expr, env)
  } else {
    suppressMessages(eval(expr, env))
  }
}

untrace_httr2 <- function(what) {
  safe_untrace(what, request)
  safe_untrace(what)
}

safe_untrace <- function(what, where = sys.frame()) {
  # If you attempt to untrace a function (1) that isn't exported from
  # whatever namespace it lives in and (2) that isn't currently traced,
  # it errors. This prevents that so that it's always safe to call `untrace`

  # untrace() and get() handle enviroments differently
  if (is.environment(where)) {
    env <- where
  } else {
    env <- environment(where)
  }
  if (inherits(try(get(what, env), silent = TRUE), "functionWithTrace")) {
    quietly(untrace(what, where = where))
  }
}

#' Serve a mock API from files
#'
#' In this context, HTTP requests attempt to load API response fixtures from
#' files. This allows test code to proceed evaluating code that expects
#' HTTP requests to return meaningful responses. Requests that do not have a
#' corresponding fixture file raise errors, like how [without_internet()]
#' does.
#'
#' `use_mock_api()` and `stop_mocking()` allow you to turn on/off request
#' mocking for more convenient use in an interactive session.
#'
#' Requests are translated to mock file paths according to several rules that
#' incorporate the request method, URL, query parameters, and body. See
#' [build_mock_url()] for details.
#'
#' File paths for API fixture files may be relative to the 'tests/testthat'
#' directory, i.e. relative to the .R test files themselves. This is the default
#' location for storing and retrieving mocks, but you can put them anywhere you
#' want as long as you set the appropriate location with [.mockPaths()].
#'
#' @param expr Code to run inside the mock context
#' @return `with_mock_api()` returns the result of `expr`. `use_mock_api()` and
#'  `stop_mocking()` return nothing.
#' @export
#' @examples
#' library(httr2)
#' with_mock_api({
#'   # There are no mocks recorded in this example, so catch this request with
#'   # expect_GET()
#'   expect_GET(
#'     request("https://cran.r-project.org") %>% req_perform(),
#'     "https://cran.r-project.org"
#'   )
#'   # For examples with mocks, see the tests and vignettes
#' })
with_mock_api <- function(expr) {
  use_mock_api()
  on.exit(stop_mocking())
  eval.parent(expr)
}

#' @rdname with_mock_api
#' @export
use_mock_api <- function() {
  options(httr2_mock = mock_request)
  # trace req_body_apply to close the file connection that it creates when the
  # body inherits form_file
  trace_httr2(
    "req_body_apply",
    exit = quote(if (exists("con")) close(con))
  )
  invisible()
}

#' @rdname with_mock_api
#' @export
stop_mocking <- function() {
  options(httr2_mock = NULL)
  untrace_httr2("req_body_apply")
  invisible()
}

mock_request <- function(req) {
  f <- build_mock_url(get_current_redactor()(req))
  mockfile <- find_mock_file(f)
  if (!is.null(mockfile)) {
    return(load_response(mockfile, req))
  }
  # Else: fail.
  # For ease of debugging if a file isn't found, include it in the
  # error that gets printed.
  req$mockfile <- f
  return(stop_request(req))
}

#' Go through mock paths to find the local mock file location
#'
#' @param file A file path, as generated by [build_mock_url()].
#' @return A path to a file that exists, or `NULL` if none found.
#' @keywords internal
#' @export
find_mock_file <- function(file) {
  for (path in .mockPaths()) {
    # Look for files of any .extension in the indicated directory,
    # be they .R, .json, ...
    mp <- file.path(path, file)
    if (file.exists(mp) && !dir.exists(mp) && grepl("\\.", basename(mp))) {
      # With write_disk() downloading, 'file' may reference a specific
      # file and include the extension .R-FILE. So if that file exists,
      # no need to search for it. Just return it.
      return(mp)
    }
    # Turn the basename into a regular expression that will match it (and
    # only it) with any .extension
    mockbasename <- paste0("^\\Q", basename(mp), "\\E.[[:alnum:]]*$")
    mockfiles <- dir(dirname(mp),
      pattern = mockbasename, all.files = TRUE,
      full.names = TRUE
    )
    # Remove directories
    mockfiles <- setdiff(mockfiles, list.dirs(dirname(mp), full.names = TRUE))
    if (length(mockfiles)) {
      # TODO: check for length > 1
      return(mockfiles[1])
    }
  }
  return(NULL)
}

#' @importFrom utils tail
#' @importFrom httr2 response
load_response <- function(file, req) {
  verbose_message("Reading ", normalizePath(file))
  ext <- tail(unlist(strsplit(file, ".", fixed = TRUE)), 1)
  if (ext == "R") {
    # It's a full "response". Source it, and if it is from httr, adapt it
    adapt_httr_response(source(file)$value)
  } else if (ext %in% names(EXT_TO_CONTENT_TYPE)) {
    response(
      url = req$url,
      method = get_request_method(req),
      headers = list(`Content-Type` = EXT_TO_CONTENT_TYPE[[ext]]),
      status_code = 200L,
      body = readBin(file, "raw", n = file.size(file))
    )
  } else if (ext == "204") {
    response(
      url = req$url,
      method = get_request_method(req),
      status_code = 204L
    )
  } else {
    stop("Unsupported mock file extension: ", ext, call. = FALSE)
  }
}

adapt_httr_response <- function(resp) {
  # Restore httr2 1.0.0 cache
  resp$cache <- new.env(parent = emptyenv())
  
  if (inherits(resp, "httr2_response")) {
    return(resp)
  }
  stopifnot(inherits(resp, "response"))

  class(resp) <- "httr2_response"
  class(resp$headers) <- "httr2_headers"
  resp$body <- resp$content
  resp$content <- resp$all_headers <- resp$cookies <- resp$times <- NULL
  resp
}

test_that("safe_untrace makes mocking not error if not already traced", {
  expect_error(use_mock_api(), NA)
  expect_error(stop_mocking(), NA)
  expect_error(stop_mocking(), NA)
})

test_that("safe_untrace makes recording not error if not already traced", {
  expect_error(start_capturing(), NA)
  expect_error(stop_capturing(), NA)
  expect_error(stop_capturing(), NA)
})

test_that("quietly muffles messages, conditional on httptest2.debug.trace", {
  expect_message(quietly(message("A message!")), NA)
  options(httptest2.debug.trace = TRUE)
  on.exit(options(httptest2.debug.trace = NULL))
  expect_message(quietly(message("A message!")), "A message!")
})

this_req <- request("http://httpbin.not/get") %>%
  req_body_json(list(test = TRUE)) %>%
  req_method("POST")

# Most of the behavior of these functions are tested elsewhere
# These tests are just for the non-standard behavior

with_mock_api({
  test_that("expect_request with mock API prints expected mock file if not found", {
    expect_error(
      req_perform(this_req),
      paste(
        "An unexpected request was made:",
        'POST http://httpbin.not/get {"test":true}',
        "Expected mock file: httpbin.not/get-f171f3-POST.*",
        sep = "\n"
      ),
      class = "httptest2_request",
      fixed = TRUE
    )
  })

  test_that("Request body and query are appended to the mock path", {
    expect_error(
      request("https://test.api/object2?d=1") %>%
        req_method("PATCH") %>%
        req_body_json(list(arg = 45)) %>%
        req_perform(),
      paste(
        "An unexpected request was made:",
        'PATCH https://test.api/object2?d=1 {"arg":45}',
        "Expected mock file: test.api/object2-899b0e-3d8d62-PATCH.*",
        sep = "\n"
      ),
      class = "httptest2_request",
      fixed = TRUE
    )
  })

  test_that("expect_VERB when no request is made", {
    suppressWarnings({
      # warnings are suppressed because of some spurious thing in testthat ... handling, will report issue
      expect_failure(
        expect_POST("just a string"),
        "No request was made"
      )
      # Actual errors are passed through
      expect_error(
        expect_failure(
          expect_POST(stop("NOTAREQUEST"))
        ),
        "NOTAREQUEST"
      )
    })
  })

  test_that("expect_no_request", {
    expect_no_request(rnorm(5))
    expect_failure(
      expect_no_request(
        request("http://httpbin.not/get") %>% req_perform()
      ),
      "An unexpected request was made"
    )
    # Make sure we aren't swallowing all errors
    expect_error(
      expect_no_request(stop("NOTAREQUEST")),
      "NOTAREQUEST"
    )
  })
})

without_internet({
  test_that("expect_request without_internet", {
    expect_failure(
      expect_POST(
        req_perform(this_req),
        "http://httpbin.not/get",
        '{"test":false}'
      ),
      paste(
        "An unexpected request was made:",
        'Actual:   POST http://httpbin.not/get {"test":true}',
        'Expected: POST http://httpbin.not/get {"test":false}',
        sep = "\n  "
      ),
      fixed = TRUE
    )
    # Error messages are different without internet (no mock file)
    expect_error(
      req_perform(this_req),
      paste(
        "An unexpected request was made:",
        'POST http://httpbin.not/get {"test":true}',
        sep = "\n"
      ),
      class = "httptest2_request",
      fixed = TRUE
    )
  })
})

without_internet({
  test_that("without_internet throws error on all verbs", {
    expect_GET(
      request(httpbin$url("/get")) %>% req_perform(),
      httpbin$url("/get")
    )
    expect_PUT(
      request(httpbin$url("/get")) %>% req_method("PUT") %>% req_perform(),
      httpbin$url("/get")
    )
    expect_POST(
      request(httpbin$url("/get")) %>% req_method("POST") %>% req_perform(),
      httpbin$url("/get")
    )
    expect_PATCH(
      request(httpbin$url("/get")) %>% req_method("PATCH") %>% req_perform(),
      httpbin$url("/get")
    )
    expect_DELETE(
      request(httpbin$url("/get")) %>% req_method("DELETE") %>% req_perform(),
      httpbin$url("/get")
    )
  })

  test_that("without_internet includes request body in message", {
    this_req <- request(httpbin$url("/get")) %>%
      req_body_json(list(test = TRUE))
    expect_PUT(
      this_req %>%
        req_method("PUT") %>%
        req_perform(),
      httpbin$url("/get"),
      '{"test":true}'
    )
    expect_POST(
      this_req %>%
        req_method("POST") %>%
        req_perform(),
      httpbin$url("/get"),
      '{"test":true}'
    )
    expect_PATCH(
      this_req %>%
        req_method("PATCH") %>%
        req_perform(),
      httpbin$url("/get"),
      '{"test":true}'
    )
  })

  test_that("max.print option", {
    options(httptest2.max.print = 3)
    on.exit(options(httptest2.max.print = NULL))
    expect_PUT(
      request(httpbin$url("/get")) %>%
        req_body_json(list(test = TRUE)) %>%
        req_method("PUT") %>%
        req_perform(),
      httpbin$url("/get"),
      '\\{"t$',
      fixed = FALSE # To use the regular expression $ to match the end
    )
  })

  test_that("without_internet respects query params", {
    expect_GET(
      request(httpbin$url("/get")) %>%
        req_url_query(test = "a phrase", two = 3) %>%
        req_perform(),
      paste0(httpbin$url("/get"), "?test=a%20phrase&two=3")
    )
  })
})

test_that("block_requests()", {
  block_requests()
  on.exit(stop_mocking())
  expect_GET(
    request(httpbin$url("/get")) %>% req_perform(),
    httpbin$url("/get")
  )
})

with_mock_api({
  test_that("Can load an object and file extension is added", {
    a <- request("https://test.api/") %>% req_perform()
    expect_identical(resp_body_json(a), list(value = "https://test.api/object1/"))
    b <- request(resp_body_json(a)$value) %>% req_perform()
    expect_identical(resp_body_json(b), list(object = TRUE))
  })
  test_that("GET with query", {
    obj <- request("https://test.api/object1/") %>%
      req_url_query(a = 1) %>%
      req_perform()
    expect_equal(
      resp_body_json(obj),
      list(query = list(a = 1), mocked = "yes"),
      list_as_map = TRUE
    )
  })
  test_that("GET with special characters", {
    a <- request("https://test.api/x(y='1',z='2')") %>% req_perform()
    expect_identical(resp_body_json(a), list(object = TRUE))
  })
  test_that("GET files that don't exist errors", {
    expect_GET(
      request("https://test.api/NOTAFILE/") %>% req_perform(), 
      "https://test.api/NOTAFILE/"
    )
    expect_GET(
      request("https://test.api/NOTAFILE/") %>%
        req_url_query(a = 1) %>%
        req_perform(),
      "https://test.api/NOTAFILE/?a=1"
    )
  })
  test_that("POST method reads from correct file", {
    b <- request("https://test.api/object1") %>%
      req_method("POST") %>%
      req_perform()
    expect_identical(resp_body_json(b), list(method = "POST"))
    b2 <- request("https://test.api/object1") %>%
      req_method("POST") %>%
      req_headers(
        Accept = "application/json",
        "Content-Type" = "application/json"
      ) %>%
      req_perform()
    expect_identical(resp_body_json(b2), list(method = "POST"))
  })
  test_that("Request body is appended to mock file path", {
    p <- request("https://test.api/object1") %>%
      req_body_raw('{"a":1}', type = "application/json") %>%
      req_headers(
        Accept = "application/json"
      ) %>%
      req_perform()
    expect_identical(resp_body_json(p), list(content = TRUE))
    expect_error(
      request("https://test.api/object1") %>%
        req_body_raw('{"b":2}', type = "application/json") %>%
        req_headers(
          Accept = "application/json"
        ) %>%
        req_perform(),
      "api/object1-3e8d9a-POST."
    )
  })
  test_that("max.print option", {
    options(httptest2.max.print = 3)
    on.exit(options(httptest2.max.print = NULL))
    expect_PUT(
      request("http://httpbin.not/get") %>%
        req_method("PUT") %>%
        req_body_json(list(test = TRUE)) %>%
        req_perform(),
      "http://httpbin.not/get",
      '{"t '
    )
  })
  test_that("Request body and query", {
    expect_PATCH(
      request("https://test.api/object2?d=1") %>%
        req_method("PATCH") %>%
        req_body_json(list(arg = 45)) %>%
        req_perform(),
      'https://test.api/object2?d=1 {"arg":45}'
    )
  })
  test_that("Other verbs error too", {
    expect_PUT(
      request("https://test.api/") %>%
        req_method("PUT") %>%
        req_perform(),
      "https://test.api/"
    )
    expect_PATCH(
      request("https://test.api/") %>%
        req_method("PATCH") %>%
        req_perform(),
      "https://test.api/"
    )
    expect_POST(
      request("https://test.api/") %>%
        req_method("POST") %>%
        req_perform(),
      "https://test.api/"
    )
    expect_POST(
      request("https://test.api/") %>%
        req_body_raw('{"arg":true}') %>%
        req_perform(),
      "https://test.api/",
      '{"arg":true}'
    )
    expect_DELETE(
      request("https://test.api/") %>%
        req_method("DELETE") %>%
        req_perform(),
      "https://test.api/"
    )
  })

  test_that("mock API with http:// URL, not file path", {
    expect_GET(
      request("http://httpbin.not/get") %>% req_perform(),
      "http://httpbin.not/get"
    )
    expect_GET(
      request("https://httpbin.not/get") %>% req_perform(),
      "https://httpbin.not/get"
    )
    expect_identical(
      resp_body_json(request("http://example.com/get") %>% req_perform()),
      list(loaded = TRUE)
    )
  })

  test_that("Mock GET with non-JSON", {
    dick <- request("http://example.com/html") %>% req_perform()
    expect_true(grepl("Melville", resp_body_string(dick)))
  })

  test_that("POST with all body types", {
    r <- request("http://httpbin.not/post") %>% req_method("POST")
    expect_POST(
      r %>%
        req_body_raw("A simple text string") %>%
        req_perform(),
      "http://httpbin.not/post",
      "A simple text string "
    )
    expect_POST(
      r %>%
        req_body_form(x = "A simple text string") %>%
        req_perform(),
      "http://httpbin.not/post",
      "x=A%20simple%20text%20string "
    )
    expect_POST(
      r %>%
        req_body_json(data = list(x = "A simple text string")) %>%
        req_perform(),
      "http://httpbin.not/post",
      '{"x":"A simple text string"} '
    )
    file_to_upload <- tempfile()
    cat("testing", file = file_to_upload)
    expect_POST(
      r %>%
        req_body_file(file_to_upload) %>%
        req_perform(),
      "http://httpbin.not/post",
      "File: ae2b1fca515949e5d54fb22b8ed95575"
    )

    expect_POST(
      r %>%
        req_body_multipart(
          a = curl::form_file(file_to_upload),
          b = curl::form_data("strings")
        ) %>%
        req_perform(),
      "http://httpbin.not/post",
      "Multipart form:
  a = File: ae2b1fca515949e5d54fb22b8ed95575
  b = strings"
    )
  })

  test_that("Regular expressions in expect_VERB", {
    expect_GET(request("http://example.com/1234/abcd/") %>% req_perform(),
      "http://example.com/[0-9]{4}/[a-z]{4}/",
      fixed = FALSE
    )
    expect_GET(request("http://example.com/1234/abcd/") %>% req_perform(),
      "http://EXAMPLE.com/[0-9]{4}/[a-z]{4}/",
      fixed = FALSE,
      ignore.case = TRUE
    )
    expect_POST(request("http://example.com/1234/abcd/") %>% req_method("POST") %>% req_perform(),
      "http://example.com/[0-9]{4}/[a-z]{4}/",
      fixed = FALSE
    )
    expect_DELETE(request("http://example.com/1234/abcd/") %>% req_method("DELETE") %>% req_perform(),
      "http://example.com/[0-9]{4}/[a-z]{4}/",
      fixed = FALSE
    )
  })
})

test_that("build_mock_url file path construction with character URL", {
  # GET (default) method
  file <- build_mock_url(request("http://www.test.com/api/call"))
  expect <- "www.test.com/api/call"
  expect_identical(file, expect, label = "Get method without query string")

  # GET method with query in URL
  file <- build_mock_url(request("http://www.test.com/api/call?q=1"))
  expect <- "www.test.com/api/call-a3679d"
  expect_identical(file, expect, label = "Get method with query string")

  # POST method
  file <- build_mock_url(request("http://www.test.com/api/call") %>% req_method("POST"))
  expect <- "www.test.com/api/call-POST"
  expect_identical(file, expect, "POST method without query string")

  # POST method with query in URL
  file <- build_mock_url(request("http://www.test.com/api/call?q=1") %>% req_method("POST"))
  expect <- "www.test.com/api/call-a3679d-POST"
  expect_identical(file, expect, "POST method with query string")
})

test_that("build_mock_url returns file names that are valid on all R platforms", {
  u <- "https://language.googleapis.com/v1/documents:annotateText/"
  expect_identical(
    build_mock_url(request(u)),
    "language.googleapis.com/v1/documents-annotateText"
  )
})

test_that("load_response invalid extension handling", {
  expect_error(
    load_response("foo.qwert"),
    "Unsupported mock file extension: qwert"
  )
})

test_that("mock_request code paths are covered (outside of trace)", {
  expect_s3_class(
    mock_request(list(method = "GET", url = "https://test.api/")),
    "httr2_response"
  )
  expect_s3_class(
    mock_request(list(method = "GET", url = "http://example.com/html")),
    "httr2_response"
  )
  expect_error(mock_request(list(method = "PUT", url = "api/")))
})

#' Set mocking/capturing state for a vignette
#'
#' Use `start_vignette()` to either use previously recorded responses, if they
#' exist, or capture real responses for future use.
#'
#' In a vignette or other R Markdown or Sweave document, place
#' `start_vignette()` in an R code block at the beginning,
#' before the first API request is made, and put
#' `end_vignette()` in a R code chunk at the end. You may
#' want to make those R code chunks have `echo=FALSE` in order to hide the fact
#' that you're calling them.
#'
#' As in [with_mock_dir()], the behavior changes based on the existence of the `dir` directory. The
#' first time you build the vignette, the directory won't exist yet, so it will
#' make real requests and record them inside of `dir`. On subsequent runs, the
#' mocks will be used. To record fresh responses from the server, delete the
#' `dir` directory, and the responses will be recorded again the next time the
#' vignette runs.
#'
#' If you have additional setup code that you'd like available across all of
#' your package's vignettes, put it in `inst/httptest2/start-vignette.R` in your
#' package, and it will be called in `start_vignette()` before the mock/record
#' context is set. Similarly, teardown code can go in
#' `inst/httptest2/end-vignette.R`, evaluated in `end_vignette()` after mocking
#' is stopped.
#'
#' @param dir Root file path for the mocks for this vignette. A good idea is
#' to use the file name of the vignette itself.
#' @param ... Optional arguments passed to `start_capturing()`
#' @return Nothing; called for its side effect of starting/ending
#' response recording or mocking.
#' @export
#' @seealso [start_capturing()] for how requests are recorded; [use_mock_api()]
#' for how previously recorded requests are loaded; [change_state()] for how to
#' handle recorded requests when the server state is changing;
#' `vignette("vignettes", package = "httptest2")` for an overview of all
start_vignette <- function(dir, ...) {
  # Cache the original .mockPaths so we can restore it on exit
  # And don't print messages in a vignette
  options(
    httptest2.mock.paths.old = getOption("httptest2.mock.paths"),
    httptest2.verbose.old = getOption("httptest2.verbose"),
    httptest2.verbose = FALSE
  )
  # This actually sources the files, if they exist
  find_package_functions(get_attached_packages(), "start-vignette.R")

  # Set the starting mockPath
  if (dir.exists("vignettes")) {
    # We're in the package root directory, probably running interactively
    # but we need to make sure we record/load relative to the vignette dir
    dir <- file.path("vignettes", dir)
  }
  .mockPaths(file.path(dir, "0"))
  if (dir.exists(dir)) {
    # We already have recorded, so use the fixtures
    use_mock_api()
  } else {
    # Record!
    start_capturing(...)
  }
}

#' Handle a change of server state
#'
#' In a vignette, put a call to `change_state()` before any code block that
#' makes a change on
#' the server, or rather, before any code block that might repeat the same
#' request previously done and expect a different result.
#'
#' `change_state()` works by layering a new directory on top of the existing
#' [.mockPaths()], so fixtures are recorded/loaded there, masking rather than
#' overwriting previously recorded responses for the same request. In
#' vignettes, these mock layers are subdirectories with integer names.
#'
#' @return Invisibly, the return of `.mockPaths()` with the new path added.
#' @seealso [start_vignette()];
#' `vignette("vignettes", package = "httptest2")` for an overview of all
#' @export
change_state <- function() {
  current_path <- .mockPaths()[1]
  path_segments <- unlist(strsplit(current_path, .Platform$file.sep))
  current_number <- suppressWarnings(as.numeric(path_segments[length(path_segments)]))
  if (is.na(current_number)) {
    # This path doesn't come from start_vignette/change_state
    stop(current_path, " is not valid for change_state()")
  }
  path_segments[length(path_segments)] <- current_number + 1
  new_path <- do.call(file.path, as.list(path_segments))
  .mockPaths(new_path)
}

#' @rdname start_vignette
#' @export
end_vignette <- function() {
  stop_capturing()
  stop_mocking()
  # TODO: compress recorded mocks (in case change_state() was called but
  # state didn't actually change) if we were recording

  # This actually sources the files, if they exist
  find_package_functions(get_attached_packages(), "end-vignette.R")

  # Restore original settings
  options(
    httptest2.mock.paths = getOption("httptest2.mock.paths.old"),
    httptest2.mock.paths.old = NULL,
    httptest2.verbose = getOption("httptest2.verbose.old")
  )
}

we_are_recording <- function() inherits(httr2:::req_perform, "functionWithTrace")
we_are_mocking <- function() !is.null(getOption("httr2_mock"))
path <- tempfile()

test_that("start/end_vignette with recording (dir does not exist yet)", {
  start_vignette(path)
  expect_true(we_are_recording())
  expect_false(we_are_mocking())
  expect_identical(.mockPaths()[1], file.path(path, "0"))
  # Test that state_change bumps the mockPath
  change_state()
  expect_identical(.mockPaths()[1], file.path(path, "1"))
  expect_identical(.mockPaths()[2], file.path(path, "0"))
  end_vignette()
  expect_false(we_are_recording())
  expect_false(we_are_mocking())
  expect_false(any(grepl(path, .mockPaths()[1], fixed = TRUE)))
})

test_that("start/end_vignette with mocking (dir exists)", {
  on.exit(options(httptest2.verbose = NULL))
  dir.create(path)
  start_vignette(path)
  expect_false(we_are_recording())
  expect_true(we_are_mocking())
  end_vignette()
  expect_false(we_are_recording())
  expect_false(we_are_mocking())
})

test_that("start_vignette puts path in vignettes dir, if exists", {
  d <- tempfile()
  dir.create(file.path(d, "vignettes"), recursive = TRUE)
  old <- setwd(d)
  on.exit(setwd(old))

  start_vignette("testing")
  expect_identical(.mockPaths()[1], file.path("vignettes", "testing", "0"))
  end_vignette()
})

test_that("start/end_vignette calls inst/httptest/vignette-start/end.R", {
  lib <- install_testpkg("testpkg")
  library(testpkg, lib.loc = lib)
  on.exit(detach("package:testpkg", unload = TRUE))
  expect_false(getOption("testpkg.start.vignette", FALSE))
  start_vignette(path)
  expect_true(getOption("testpkg.start.vignette"))

  end_vignette()
  expect_false(getOption("testpkg.start.vignette", FALSE))
})

test_that("change_state validation", {
  with_mock_path("foo", {
    expect_error(change_state(), "foo is not valid for change_state()")
  })
})

reset_redactors()

#' Record API responses as mock files
#'
#' `capture_requests()` is a context that collects the responses from requests
#' you make and stores them as mock files. This enables you to perform a series
#' of requests against a live server once and then build your test suite using
#' those mocks, running your tests in [with_mock_api()].
#'
#' `start_capturing()` and `stop_capturing()` allow you to turn on/off request
#' recording for more convenient use in an interactive session.
#'
#' Recorded responses are written out as plain-text files. By storing fixtures
#' as plain-text files, you can
#' more easily confirm that your mocks look correct, and you can more easily
#' maintain them without having to re-record them. If the API changes subtly,
#' such as when adding an additional attribute to an object, you can just touch
#' up the mocks.
#'
#' If the response has status `200 OK` and the `Content-Type`
#' maps to a supported file extension---currently `.json`,
#' `.html`, `.xml`, `.txt`, `.csv`, and `.tsv`---just the response body will be
#' written out, using the appropriate extension. `204 No Content` status
#' responses will be stored as an empty file with extension `.204`. Otherwise,
#' the response will be written as a `.R` file containing syntax that, when
#' executed, recreates the `httr2_response` object.
#'
#' Files are saved to the first directory in [.mockPaths()], which if not
#' otherwise specified is either "tests/testthat" if it exists
#' (as it should if you are in the root directory of your package),
#' else the current working directory.
#' If you have trouble when recording responses, or are unsure where the files
#' are being written, set `options(httptest2.verbose = TRUE)` to print a message
#' for every file that is written containing the absolute path of the file.
#'
#' @param expr Code to run inside the context
#' @param simplify logical: if `TRUE` (default), plain-text responses with status 200
#' will be written as just the text of the response body. In all other cases,
#' and when `simplify` is `FALSE`, the `httr2_response` object will be written out to
#' a .R file using [base::dput()].
#' @return `capture_requests()` returns the result of `expr`. `start_capturing()`
#' invisibly returns the destination directory.
#' `stop_capturing()` returns nothing; it is called for its side effects.
#' @examplesIf FALSE
#' # Setup so that our examples clean up after themselves
#' tmp <- tempfile()
#' .mockPaths(tmp)
#' on.exit(unlink(tmp, recursive = TRUE))
#'
#' library(httr2)
#' capture_requests({
#'   request("http://httpbin.org/get") %>% req_perform()
#'   request("http://httpbin.org/response-headers") %>%
#'     req_headers(`Content-Type` = "application/json") %>%
#'     req_perform()
#' })
#' # Or:
#' start_capturing()
#' request("http://httpbin.org/get") %>% req_perform()
#' request("http://httpbin.org/response-headers") %>%
#'   req_headers(`Content-Type` = "application/json") %>%
#'   req_perform()
#' stop_capturing()
#' @export
#' @seealso [build_mock_url()] for how requests are translated to file paths.
#' And see `vignette("redacting", package = "httptest2")`
#' for details on how to prune sensitive content from responses when recording.
capture_requests <- function(expr, simplify = TRUE) {
  start_capturing(simplify)
  on.exit(stop_capturing())
  eval.parent(expr)
}

#' @rdname capture_requests
#' @export
start_capturing <- function(simplify = TRUE) {
  # Use "substitute" so that `simplify` is inserted. Code remains quoted.
  req_tracer <- substitute(
    {
      # Get the value returned from the function, and sanitize it
      redactor <- httptest2::get_current_redactor()
      if (exists("mock_resp") && !is.null(mock_resp)) {
        # We're mocking and returning early
        resp <- mock_resp
      }
      if (!exists("resp") || inherits(resp, "error")) {
        # req_perform() catches request errors as an "error" object
        # See httr2:::is_error()
        # Also possible that resp doesn't exist if the function errors eariler
        warning(
          "Request errored; no captured response file saved",
          call. = FALSE
        )
      } else {
        httptest2::save_response(
          redactor(resp),
          file = httptest2::build_mock_url(redactor(req)),
          simplify = simplify
        )
      }
    },
    list(simplify = simplify)
  )
  trace_httr2("req_perform", exit = req_tracer)
  invisible(.mockPaths())
}

#' Write out a captured response
#'
#' @param response An `httr2_response` object
#' @param file String file path to write to
#' @param simplify logical: if `TRUE` (default), JSON responses with status 200
#' and a supported `Content-Type`
#' will be written as just the text of the response body. In all other cases,
#' and when `simplify` is `FALSE`, the `httr2_response` object will be written
#' out to a .R file using [base::dput()].
#' @return The character file name that was written out
#' @export
#' @keywords internal
#' @importFrom jsonlite prettify
#' @importFrom httr2 resp_content_type resp_status resp_body_string
save_response <- function(response, file, simplify = TRUE) {
  # Track separately the actual full path we're going to write to
  dst_file <- file.path(.mockPaths()[1], file)
  mkdir_p(dst_file)

  # Get the Content-Type
  ct <- resp_content_type(response)
  status <- resp_status(response)
  if (simplify && status == 200 && ct %in% names(CONTENT_TYPE_TO_EXT)) {
    if (length(response$body)) {
      cont <- resp_body_string(response)
      if (ct == "application/json") {
        # Prettify
        cont <- prettify(cont)
      }
    } else {
      cont <- character()
    }
    dst_file <- paste(dst_file, CONTENT_TYPE_TO_EXT[[ct]], sep = ".")
    cat_wb(cont, file = dst_file)
  } else if (simplify && status == 204) {
    # "touch" a file with extension .204
    dst_file <- paste0(dst_file, ".204")
    cat_wb("", file = dst_file)
  } else {
    # Dump an object that can be sourced

    # Change the file extension to .R
    dst_file <- paste0(dst_file, ".R")
    file <- paste0(file, ".R")

    # If content is text, rawToChar it and dput it as charToRaw(that)
    # so that it loads correctly but is also readable
    text_types <- c(
      "application/json",
      "application/x-www-form-urlencoded", "application/xml",
      "text/csv", "text/html", "text/plain",
      "text/tab-separated-values", "text/xml"
    )
    if (is.raw(response$body) && ct %in% text_types && length(response$body)) {
      cont <- resp_body_string(response)
      response$body <- substitute(charToRaw(cont))
    } else if (inherits(response$body, "httr2_path")) {
      # Copy real file and substitute the response$content "path".
      downloaded_file <- paste0(dst_file, "-FILE")
      file.copy(response$body, downloaded_file)
      file <- paste0(file, "-FILE")
      response$body <- substitute(structure(find_mock_file(file),
        class = "httr2_path"
      ))
    }
    # Needed for httr2 1.0.0
    response$cache <- quote(new.env(parent = emptyenv()))
    response$request <- NULL

    f <- file(dst_file, "wb", encoding = "UTF-8")
    on.exit(close(f))
    dput(response, file = f)
  }
  verbose_message("Writing ", normalizePath(dst_file))
  return(dst_file)
}

#' @rdname capture_requests
#' @export
stop_capturing <- function() {
  untrace_httr2("req_perform")
  invisible()
}

mkdir_p <- function(filename) {
  # Recursively create the directories so that we can write this file.
  # If they already exist, do nothing.
  # Like mkdir -p path

  dir.create(dirname(filename), showWarnings = FALSE, recursive = TRUE)
}

cat_wb <- function(x, file, ...) {
  # For cleaning up CRLF issues on Windows, write to a file connection in binary mode
  f <- file(file, "wb", encoding = "UTF-8")
  on.exit(close(f))
  cat(enc2utf8(x), file = f, ...)
}

verbose_message <- function(...) {
  if (isTRUE(getOption("httptest2.verbose", FALSE))) {
    message(...)
  }
}

d <- tempfile()
dl_file <- tempfile()
webp_file <- tempfile()

# The webfake URL will be something like 127.0.0.1:port, and port may vary
# so the mock paths will be different every time it runs
httpbin_mock_url <- build_mock_url(request(httpbin$url("")))

test_that("We can record a series of requests (a few ways)", {
  with_mock_path(d, {
    capture_requests({
      # <<- assign these so that they're available in the next test_that too
      r1 <<- request(httpbin$url("/get")) %>% req_perform()
      r2 <<- request(httpbin$url("")) %>% req_perform()
      r3 <<- request(httpbin$url("/status/418")) %>%
        req_error(is_error = ~FALSE) %>%
        req_perform()
      r4 <<- request(httpbin$url("/put")) %>%
        req_method("PUT") %>%
        req_perform()
    })
    # Now do some with start/stop
    start_capturing()
    r5 <<- request(httpbin$url("/response-headers")) %>%
      req_url_query(`Content-Type` = "application/json") %>%
      req_perform()
    # Now, some requests that write to disk
    r6 <<- request(httpbin$url("/anything")) %>%
      req_perform(path = dl_file)
    r7 <<- request(httpbin$url("/image/webp")) %>%
      req_perform(path = webp_file)
    r8 <<- request(httpbin$url("/status/202")) %>% req_perform()
    r9 <<- request(httpbin$url("/status/200")) %>% req_perform()
    stop_capturing()
  })

  # If these were recorded against httpbin.org, it would look like this:
  expected_files <- c(
    "httpbin.org.html", # it's HTML, and we now support that simplified
    "httpbin.org/anything.json",
    "httpbin.org/get.json",
    "httpbin.org/image/webp.R", # Not a simplifiable format, so .R
    "httpbin.org/image/webp.R-FILE", # The `write_disk` location
    "httpbin.org/put-PUT.json", # Not a GET, but returns 200
    "httpbin.org/response-headers-ac4928.json",
    "httpbin.org/status/200.txt", # empty 200 response "text/plain", so .txt
    "httpbin.org/status/202.R", # Not 200 response, so .R
    "httpbin.org/status/418.R" # Not 200 response, so .R
  )
  # But since we don't use httpbin anymore, they're in the localhost-port dir
  expected_files <- sub("httpbin.org", httpbin_mock_url, expected_files)
  expect_identical(sort(dir(d, recursive = TRUE)), expected_files)

  # Test the contents of the .R files
  teapot <- source(file.path(d, httpbin_mock_url, "status", "418.R"))$value
  expect_s3_class(teapot, "httr2_response")
  expect_identical(resp_status(teapot), 418L)
  expect_false("request" %in% names(teapot))
  # Make sure that our .html file has HTML
  expect_true(any(grepl(
    "</body>",
    suppressWarnings(readLines(file.path(d, paste0(httpbin_mock_url, ".html"))))
  )))
})

test_that("We can then load the mocks it stores", {
  # Look for mocks in our temp dir
  with_mock_path(d, {
    # Because the place we wrote out the file in our real request might not
    # naturally be in our mock directory, assume that that file doesn't exist
    # when we load our mocks.
    content_r6 <<- resp_body_json(r6)
    file.remove(dl_file)
    content_r7 <<- resp_body_raw(r7)
    file.remove(webp_file)

    mock_dl_file <- tempfile()
    mock_webp_file <- tempfile()
    with_mock_api({
      m1 <- request(httpbin$url("/get")) %>% req_perform()
      m2 <- request(httpbin$url("")) %>% req_perform()
      m3 <- request(httpbin$url("/status/418")) %>%
        req_error(is_error = ~FALSE) %>%
        req_perform()
      m4 <- request(httpbin$url("/put")) %>%
        req_method("PUT") %>%
        req_perform()
      m5 <- request(httpbin$url("/response-headers")) %>%
        req_url_query(`Content-Type` = "application/json") %>%
        req_perform()
      m6 <- request(httpbin$url("/anything")) %>%
        req_perform(path = mock_dl_file)
      m7 <- request(httpbin$url("/image/webp")) %>%
        req_perform(path = mock_webp_file)
      m8 <- request(httpbin$url("/status/202")) %>% req_perform()
      m9 <- request(httpbin$url("/status/200")) %>% req_perform()
    })
  })
  expect_identical(resp_body_json(m1), resp_body_json(r1))
  # Compare the HTML as text because the parsed HTML (XML) object has a
  # C pointer that is different between the two objects.
  expect_identical(
    enc2native(resp_body_string(m2)),
    enc2native(resp_body_string(r2))
  )

  expect_true(grepl("</body>", resp_body_string(m2)))
  expect_identical(resp_body_string(m3), resp_body_string(r3))
  expect_identical(resp_body_json(m4), resp_body_json(r4))
  expect_identical(resp_body_json(m5), resp_body_json(r5))
  expect_identical(resp_body_json(m6), content_r6)
  expect_identical(resp_body_raw(m7), content_r7)
  expect_equal(resp_status(m8), 202)
  expect_equal(resp_status(m9), 200)
  expect_equal(resp_content_type(m9), "text/plain")
  expect_false(resp_has_body(m9))
})

test_that("write_disk mocks can be reloaded even if the mock directory moves", {
  # This is an edge case caught because `crunch` package puts fixtures in
  # `inst/`, so you record to one place but when you read them from the
  # installed package, it's a different directory.
  d2 <- tempfile()
  dir.create(file.path(d2, httpbin_mock_url, "image"), recursive = TRUE)
  for (f in c("webp.R", "webp.R-FILE")) {
    file.rename(
      file.path(d, httpbin_mock_url, "image", f),
      file.path(d2, httpbin_mock_url, "image", f)
    )
  }
  with_mock_path(d2, {
    with_mock_api({
      m7b <- request(httpbin$url("/image/webp")) %>%
        req_perform(tempfile())
    })
  })
  expect_identical(resp_body_raw(m7b), content_r7)
})

with_mock_api({
  d2 <- tempfile()
  test_that("Recording requests even with the mock API", {
    capture_while_mocking(path = d2, {
      request("http://example.com/get/") %>% req_perform()
      request("https://test.api/object1/") %>% req_perform()
      request("http://httpbin.not/status/204/") %>% req_perform()
    })
    expect_setequal(
      dir(d2, recursive = TRUE),
      c("example.com/get.json", "test.api/object1.json", "httpbin.not/status/204.204")
    )
    expect_identical(
      readLines(file.path(d2, "example.com/get.json")),
      readLines("example.com/get.json")
    )
  })

  test_that("Loading 204 response status recorded with simplify=TRUE", {
    original <- request("http://httpbin.not/status/204/") %>% req_perform()
    expect_length(original$body, 0)
    expect_length(
      readLines(file.path(d2, "httpbin.not/status/204.204")),
      0
    )
    with_mock_path(d2,
      {
        mocked <- request("http://httpbin.not/status/204/") %>% req_perform()
        expect_length(mocked$body, 0)
      },
      replace = TRUE
    )
  })

  d3 <- tempfile()
  test_that("Using simplify=FALSE (and setting .mockPaths)", {
    with_mock_path(d3, {
      capture_while_mocking(simplify = FALSE, {
        request("http://example.com/get/") %>% req_perform()
        request("https://test.api/object1/") %>% req_perform()
        request("http://httpbin.not/status/204/") %>% req_perform()
      })
    })
    expect_setequal(
      dir(d3, recursive = TRUE),
      c("example.com/get.R", "test.api/object1.R", "httpbin.not/status/204.R")
    )
    response <- source(file.path(d3, "example.com/get.R"))$value
    expect_s3_class(response, "httr2_response")
    expect_identical(
      resp_body_json(response),
      resp_body_json(request("http://example.com/get/") %>% req_perform())
    )
  })

  test_that("Recorded JSON is prettified", {
    expect_length(
      readLines(file.path(d2, "example.com/get.json")),
      3L
    )
    skip("TODO: prettify when simplify=FALSE")
    response <- readLines(file.path(d3, "api/object1.R"))
  })

  test_that("Using options(httptest2.verbose = TRUE) works", {
    d4 <- tempfile()
    old <- options(httptest2.verbose = TRUE)
    on.exit(options(old))
    with_mock_path(d4, {
      capture_while_mocking(
        expect_message(
          expect_message(
            request("http://example.com/get/") %>% req_perform(),
            "Writing .*example.com.get.json"
          ),
          "Reading .*example.com.get.json"
        )
      )
    })
    expect_setequal(
      dir(d4, recursive = TRUE),
      c("example.com/get.json")
    )
    expect_identical(
      readLines(file.path(d4, "example.com/get.json")),
      readLines("example.com/get.json")
    )
  })
})

test_that("If the httr2 request function exits with an error, capture_requests warns", {
  capture_requests({
    with_mock(
      function(req) stop("Error!"),
      expect_warning(
        expect_error(
          request("http://httpbin.not/get") %>% req_perform()
        ),
        "Request errored; no captured response file saved"
      )
    )
  })
})

Sys.setlocale("LC_COLLATE", "C") # What CRAN does
set.seed(999)
options(
  warn = 1,
  httptest2.debug.trace = FALSE
)

library(httr2, warn.conflicts = FALSE)

httpbin <- webfakes::local_app_process(
  webfakes::httpbin_app(),
  .local_envir = testthat::teardown_env()
)

capture_while_mocking <- function(..., path) {
  with_mock_path(path, {
    # We'll write to `path` but read from wherever was set before
    tracer <- quote({
      .mockPaths <- function() getOption("httptest2.mock.paths")[-1]
    })
    with_trace("find_mock_file",
      where = with_mock_api, tracer = tracer,
      expr = capture_requests(...)
    )
  })
}

with_redactor <- function(x, ...) {
  old <- getOption("httptest2.redactor")
  old.pkgs <- getOption("httptest2.redactor.packages")
  set_redactor(x)
  on.exit({
    set_redactor(old)
    options(httptest2.redactor.packages = old.pkgs)
  })
  eval.parent(...)
}

reset_redactors <- function() {
  options(
    httptest2.redactor = NULL,
    httptest2.redactor.packages = NULL
  )
}

install_testpkg <- function(pkg, lib = tempfile()) {
  dir.create(lib)
  tools::Rcmd(c("INSTALL", "testpkg", paste0("--library=", shQuote(lib))),
    stdout = NULL, stderr = NULL
  )
  return(lib)
}

expect_redactor <- function(expr) {
  expect_identical(names(formals(expr)), "response")
}

test_that("prepare_redactor: function", {
  expect_identical(prepare_redactor(redact_cookies), redact_cookies)
})

multiredact <- list(redact_cookies, ~ redact_headers(., "X-Header"))
test_that("prepare_redactor: list/multiple", {
  expect_redactor(prepare_redactor(multiredact))
  expect_identical(prepare_redactor(multiredact[1]), redact_cookies)
})

test_that("prepare_redactor: NULL for no redacting", {
  expect_identical(prepare_redactor(NULL), force)
})

test_that("prepare_redactor: garbage", {
  expect_error(
    prepare_redactor("foo"),
    "Redactor must be a function or list of functions"
  )
})

test_that("get_current_redactor edge cases", {
  options(httptest2.redactor = NULL)
  expect_identical(get_current_redactor(), redact_cookies)
  options(httptest2.redactor.packages = "NOTAPACKAGE")
  expect_identical(get_current_redactor(), redact_cookies)
})

with_mock_api({
  test_that("Reading redactors from within a package (and install that package)", {
    newmocks <- tempfile()
    # Also show verbose messaging of redactor
    options(httptest2.verbose = TRUE)
    expect_message(
      expect_message(
        expect_message(
          capture_while_mocking(path = newmocks, {
            # Install the "testpkg" to a temp lib.loc _after_ we've
            # already started recording
            lib <- install_testpkg("testpkg")
            library(testpkg, lib.loc = lib)
            expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))
            r <- request("http://example.com/get") %>% req_perform()
          }),
          paste0("Using redact.R from ", dQuote("testpkg"))
        ),
        "Writing .*example.com.get.json"
      ),
      "Reading .*example.com.get.json"
    )
    options(httptest2.verbose = FALSE)
    with_mock_path(newmocks, {
      r2 <- request("http://example.com/get") %>% req_perform()
    })
    # The resulting mock content is what we injected into it from testpkg
    expect_identical(resp_body_json(r2), list(fake = TRUE))
  })

  test_that("set_redactor(NULL) to override default (and loaded packages)", {
    expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))
    # Great, but let's kill it when we're done
    on.exit(detach("package:testpkg", unload = TRUE))
    newmocks2 <- tempfile()
    with_redactor(
      NULL,
      capture_while_mocking(simplify = FALSE, path = newmocks2, {
        a <- request("http://example.com/login") %>%
          req_body_json(list(username = "password")) %>%
          req_perform()
      })
    )
    # The auth token, which would have been removed (see test-redact.R),
    # is present in the file we wrote because we set NULL as redactor
    expect_true(any(grepl(
      "token=12345",
      readLines(file.path(newmocks2, "example.com", "login-712027-POST.R"))
    )))
  })

  test_that("Loading a package with pkgload (devtools)", {
    newmocks3 <- tempfile()
    expect_false("testpkg" %in% names(sessionInfo()$otherPkgs))
    on.exit(pkgload::unload("testpkg"))
    capture_while_mocking(path = newmocks3, {
      pkgload::load_all("testpkg", quiet = TRUE)
      expect_true("testpkg" %in% names(sessionInfo()$otherPkgs))
      r <- request("http://example.com/get") %>% req_perform()
    })
    with_mock_path(newmocks3, {
      r2 <- request("http://example.com/get") %>% req_perform()
    })
    # The resulting mock content is what we injected into it from testpkg
    expect_identical(resp_body_json(r2), list(fake = TRUE))
  })
})

reset_redactors()

#' Use or create mock files depending on their existence
#'
#' This context will switch the [.mockPaths()] to `tests/testthat/dir`
#' (and then resets it to what it was before).
#' If the `tests/testthat/dir` folder doesn't exist, [capture_requests()] will
#' be run to create mocks.
#' If it exists, [with_mock_api()] will be run.
#' To re-record mock files, simply delete `tests/testthat/dir` and run the test.
#'
#' @param dir character string, unique folder name that will be used or created
#' under `tests/testthat/`
#' @inheritParams start_capturing
#' @param replace Logical: should `dir` replace the contents of `.mockPaths()`
#' (default) or be added in front of the existing paths? The default behavior
#' here is the opposite of `.mockPaths()` so that the tests inside of
#' `with_mock_dir()` are fully isolated.
#' @seealso `vignette("httptest2")` for usage examples.
#' @export
with_mock_dir <- function(dir, expr, simplify = TRUE, replace = TRUE) {
  if (dir.exists("tests/testthat") && !(substr(dir, 1, 1) %in% c("/", "\\"))) {
    # If we're at the top level directory of the package,
    # assume any relative paths are meant to be inside tests/testthat.
    dir <- file.path("tests", "testthat", dir)
  }
  with_mock_path(dir, replace = replace, {
    if (dir.exists(dir)) {
      # We already have recorded, so use the fixtures
      verbose_message("Using mocks found in ", dir)
      with_mock_api(expr)
    } else {
      # Record!
      verbose_message("Recording responses to ", dir)
      capture_requests(expr, simplify = simplify)
    }
  })
}

#' Use 'httptest2' in your tests
#'
#' This function adds `httptest2` to Suggests in the package DESCRIPTION and
#' loads it in `tests/testthat/setup.R`. Call it once when you're setting up
#' a new package test suite.
#'
#' The function is idempotent: if `httptest2` is already added to these files,
#' no additional changes will be made.
#'
#' @param path character path to the package
#' @return Nothing: called for file system side effects.
#' @export
use_httptest2 <- function(path = ".") {
  if (!("DESCRIPTION" %in% dir(path))) {
    stop(path, " is not an R package directory", call. = FALSE)
  }
  add_httptest2_to_desc(file.path(path, "DESCRIPTION"))
  add_httptest2_to_setup(file.path(path, "tests", "testthat", "setup.R"))
  invisible()
}

add_httptest2_to_desc <- function(file) {
  # Read DESCRIPTION, add httptest2 to Suggests if not already there

  # Hack to preserve whitespace: read it twice
  desc_fields <- colnames(read.dcf(file))
  desc <- read.dcf(file, keep.white = desc_fields)
  if (!("Suggests" %in% desc_fields)) {
    # Add a column for Suggests
    desc <- cbind(desc, matrix("", ncol = 1, dimnames = list(NULL, "Suggests")))
  }
  if (!grepl("httptest2", desc[, "Suggests"])) {
    # Add httptest2

    # Parse the list, and try to preserve the whitespace from the original
    suggested_pkgs <- unlist(strsplit(desc[, "Suggests"], ","))
    sep <- sub("^([[:blank:]\n]*).*", "\\1", suggested_pkgs)
    suggested_pkgs <- sort(c(trimws(suggested_pkgs), "httptest2"))
    extra_sep <- tail(unique(sep), 1)
    if (length(extra_sep) == 0 || nchar(extra_sep) == 0) {
      extra_sep <- " "
    }
    sep <- c(sep, extra_sep)
    desc[, "Suggests"] <- paste0(sep, suggested_pkgs, collapse = ",")

    # Msg and write
    message("Adding 'httptest2' to Suggests in DESCRIPTION")
    write.dcf(desc, file = file, keep.white = desc_fields)
  }
}

add_httptest2_to_setup <- function(file) {
  # Create tests/testthat/setup.R if it does not exist

  if (!file.exists(file)) {
    message("Creating ", file)
    message("Adding library(httptest2) to ", file)
    mkdir_p(file)
    cat("library(httptest2)\n", file = file)
    # Msg and write
  } else {
    setup_lines <- readLines(file)
    if (!any(grepl("library(httptest2)", setup_lines, fixed = TRUE))) {
      # Add "library(httptest)" to the top if it's not already there
      setup_lines <- c("library(httptest2)", setup_lines)
      # Msg and write
      message("Adding library(httptest2) to ", file)
      writeLines(setup_lines, file)
    }
  }
}

test_add_to_desc <- function(str, msg = "Adding 'httptest2' to Suggests") {
  f <- tempfile()
  cat(str, file = f)
  expect_message(add_httptest2_to_desc(f), msg)
  setNames(read.dcf(f, keep.white = "Suggests")[, "Suggests"], NULL)
}

test_that("add to desc adds Suggests if not present", {
  expect_identical(test_add_to_desc("Title: Foo"), "httptest2")
})

test_that("add to desc adds Suggests if present but empty", {
  expect_identical(test_add_to_desc("Suggests:"), "httptest2")
})

test_that("add to desc adds Suggests inline", {
  expect_identical(test_add_to_desc("Suggests: pkg"), "httptest2, pkg")
  expect_identical(
    test_add_to_desc("Suggests: pkg, alpha (>= 2.0.0)"),
    "alpha (>= 2.0.0), httptest2, pkg"
  )
})

desc_suggests_one_multiline <- "Title: Foo
Suggests:
    pkg
"
desc_suggests_two_multiline <- "Title: Foo
Suggests:
    pkg,
    alpha (>= 2.0.0)
"
desc_suggests_two_uneven <- "Title: Foo
Suggests: pkg,
    alpha (>= 2.0.0)
"

test_that("add to desc adds Suggests multiline", {
  expect_identical(
    test_add_to_desc(desc_suggests_one_multiline),
    "\n    httptest2,\n    pkg"
  )
  expect_identical(
    test_add_to_desc(desc_suggests_two_multiline),
    "\n    alpha (>= 2.0.0),\n    httptest2,\n    pkg"
  )
  expect_identical(
    test_add_to_desc(desc_suggests_two_uneven),
    "alpha (>= 2.0.0),\n    httptest2,\n    pkg"
  )
})

test_that("add to desc doesn't add if already present", {
  expect_identical(test_add_to_desc("Suggests: httptest2", msg = NA), "httptest2")
  expect_identical(
    test_add_to_desc("Suggests: pkg, httptest2", msg = NA),
    "pkg, httptest2"
  )
})

expect_added_to_setup <- function(str, msg = "Adding library\\(httptest2\\)") {
  f <- tempfile()
  cat(str, file = f)
  expect_message(add_httptest2_to_setup(f), msg)
  expect_true(any(grepl("library(httptest2)", readLines(f), fixed = TRUE)))
}

test_that("add to setup creates file if doesn't exist", {
  f <- tempfile()
  expect_false(file.exists(f))
  expect_message(
    expect_message(add_httptest2_to_setup(f), "Creating"),
    "Adding library\\(httptest2\\) to"
  )
  expect_identical(readLines(f), "library(httptest2)")
})

test_that("add to setup adds", {
  expect_added_to_setup("")
  expect_added_to_setup("library(pkg)\n")
})

test_that("add to setup doesn't duplicate", {
  expect_added_to_setup("library(httptest2)\n", msg = NA)
})

test_that("use_httptest2 integration test", {
  testpkg <- tempfile()
  dir.create(testpkg)
  expect_error(use_httptest2(testpkg), "is not an R package directory")

  desc <- file.path(testpkg, "DESCRIPTION")
  cat("Title: Foo\n", file = desc)
  setup <- file.path(testpkg, "tests", "testthat", "setup.R")
  expect_message(
    expect_message(
      expect_message(
        use_httptest2(testpkg),
        "Adding 'httptest2' to Suggests"
      ),
      "Creating "
    ),
    "Adding library\\(httptest2\\) to "
  )
  expect_identical(readLines(desc), c("Title: Foo", "Suggests: httptest2"))
  expect_identical(readLines(setup), "library(httptest2)")
  # It does nothing if you the package already uses httptest2
  expect_message(use_httptest2(testpkg), NA)
})

#' Expectations for mocked HTTP requests
#'
#' The mock contexts in `httptest2` can raise errors or messages when requests
#' are made, and those (error) messages have three
#' elements, separated by space: (1) the request
#' method (e.g. "GET"); (2) the request URL; and
#' (3) the request body, if present.
#' These verb-expectation functions look for this message shape. `expect_PUT`,
#' for instance, looks for a request message that starts with "PUT".
#'
#' @param object Code to execute that may cause an HTTP request
#' @param url character: the URL you expect a request to be made to. Default is
#' an empty string, meaning that you can just assert that a request is made with
#' a certain method without asserting anything further.
#' @param ... character segments of a request payload you expect to be included
#' in the request body, to be joined together by `paste0()`. You may also pass
#' any of the following named logical arguments, which will be passed to
#' [base::grepl()]:
#' * `fixed`: Should matching take the pattern as is or treat it as a regular
#' expression. Default: `TRUE`, and note that this default is the opposite of
#' the default in `grepl`. (The rest of the arguments follow its defaults.)
#' * `ignore.case`: Should matching be done case insensitively? Default:
#' `FALSE`, meaning matches are case sensitive.
#' * `perl`: Should Perl-compatible regular expressions be used? Default: `FALSE`
#' * `useBytes`: Should matching be done byte-by-byte rather than
#' character-by-character? Default: `FALSE`
#' @return A `testthat` 'expectation'.
#' @examples
#' library(httr2)
#' without_internet({
#'   expect_GET(
#'     request("http://httpbin.org/get") %>% req_perform(),
#'     "http://httpbin.org/get"
#'   )
#'   expect_GET(
#'     request("http://httpbin.org/get") %>% req_perform(),
#'     "http://httpbin.org/[a-z]+",
#'     fixed = FALSE # For regular expression matching
#'   )
#'   expect_PUT(
#'     request("http://httpbin.org/put") %>%
#'       req_method("PUT") %>%
#'       req_body_json(list(a = 1)) %>%
#'       req_perform(),
#'     "http://httpbin.org/put",
#'     '{"a":1}'
#'   )
#'   # Don't need to assert the request body, or even the URL
#'   expect_PUT(
#'     request("http://httpbin.org/put") %>%
#'       req_method("PUT") %>%
#'       req_body_json(list(a = 1)) %>%
#'       req_perform()
#'   )
#'   expect_no_request(rnorm(5))
#' })
#' @name expect_verb
#' @aliases expect_GET expect_POST expect_PUT expect_PATCH expect_DELETE expect_no_request
#' @export
expect_GET <- function(object, url = "", ...) {
  expect_request(object, "GET ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_POST <- function(object, url = "", ...) {
  expect_request(object, "POST ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_PATCH <- function(object, url = "", ...) {
  expect_request(object, "PATCH ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_PUT <- function(object, url = "", ...) {
  expect_request(object, "PUT ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_DELETE <- function(object, url = "", ...) {
  expect_request(object, "DELETE ", url, " ", ...)
}

#' @rdname expect_verb
#' @export
expect_no_request <- function(object, ...) {
  # No request means no httptest2_request error raised
  expect_error(object, NA, ..., class = "httptest2_request")
}

#' @importFrom testthat expect_error fail
expect_request <- function(object,
                           ...,
                           fixed = TRUE,
                           ignore.case = FALSE,
                           perl = FALSE,
                           useBytes = FALSE) {
  # PUT/POST/PATCH with no body may have trailing whitespace, so trim it
  expected <- sub(" +$", "", paste0(...))
  withCallingHandlers(
    expect_error(
      object,
      expected,
      class = "httptest2_request",
      fixed = fixed,
      ignore.case = ignore.case,
      perl = perl,
      useBytes = useBytes
    ),
    error = function(e) {
      # Distinguish between unexpected requests and real errors
      if (inherits(e, "httptest2_request")) {
        fail(paste0(
          "An unexpected request was made:\n  Actual:   ",
          e$message,
          "\n  Expected: ", expected
        ))
      } else if (inherits(e, "expectation_failure") && grepl("did not throw", e$message)) {
        # This is what expect_error() says (under testthat 3e) when there is no error
        fail("No request was made")
      } else {
        # A regular error. Re-raise it.
        stop(e)
      }
    }
  )
}

---
title: "Frequently Asked Questions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Frequently Asked Questions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Where are my mocks recorded?

By default, `with_mock_api()` will look for and `capture_requests()` will
write mocks to your package's `tests/testthat` directory, or else the current
working directory if that path does not exist. If you want to look in or
write to other places, call `.mockPaths()` to add directories to the search
path.

If you're running `capture_requests()` within a test suite in an installed package, or if you're running interactively from a different directory, the working directory may not be the same as your code repository. If you aren't sure where the files are going, set `options(httptest2.verbose = TRUE)`, and it will message the absolute path of the files as it writes them.

To change where files are being written or read from, use `.mockPaths()` (like `base::.libPaths()`) to specify a different directory.

## How do I fix "non-portable file paths"?

If you see this error in `R CMD build` or `R CMD check`, it means that there are file paths are longer than 100 characters, which can sometimes happen when you record requests. `httptest` preserves the URL structure of mocks in file paths to improve the readability and maintainability of your tests, as well as to make visible the properties of your API.
Indeed, the file-system tree view of the mock files gives a visual representation of your API. This value comes with a tradeoff: sometimes URLs can be long, and R doesn't like that.

Depending on how long your URLs are, there are a few ways to save on characters without compromising readability of your code and tests.

A good way to cut long file paths is by using a redactor: a function that alters the content of your requests and responses before mapping them to mock files. For example, if all of your API endpoints sit beneath `https://language.googleapis.com/v1/`, you could:

```r
set_redactor(function (x) {
  gsub_response(x, "https\\://language.googleapis.com/v1/", "api/")
})
```

This will replace that string in all parts of the mock file that is saved, including in the file path that is written--that is, paths will start with "api/" rather than "language.googleapis.com/v1/", saving you (in this case) 23 characters. The function will also be called when loading mocks in `with_mock_api()` so that the shortened file paths are found.

You can also provide this function in `inst/httptest2/redact.R`, and any time your package is loaded (as when you run tests or build vignettes, or when anyone else uses your package in *their* package), this function will be called automatically. See `vignette("redacting")` for more.

You may also be able to economize on other parts of the file paths. If you've recorded requests and your file paths contain long ids like "1495480537a3c1bf58486b7e544ce83d", depending on how you access the API in your code, you may be able to simply replace that id with something shorter, like "1". The mocks are just files, disconnected from a real server and API, so you can rename them and munge them as needed.

Finally, if you have your tests inside a `tests/testthat/` directory, and your fixture files inside that, you can save 9 characters by moving the fixtures up to `tests/` and setting `.mockPaths("../")`.

## Why are mocks stored as text files?

Plain-text files have several nice features, particularly relative to storing serialized (binary) R objects. You can more easily confirm that your mocks look correct, and you can more easily maintain them without having to re-record them. When you do edit them, text files are more easily handled by version-control systems like Git. Plain-text files can also have comments, so you can make notes as to why a certain fixture exists, what a particular value means, and so on, which will help the users of your package—and your future self!

By having mocks in human-readable text files, you can also more easily extend your code. APIs are living things that evolve over time, and your code that communicates with an API needs to be able to change with them. If the API adds an additional attribute to an object, no big deal: just touch up the mocks. In addition, you can future-proof your code against that kind of API change by tweaking a fixture file. In [this example](https://github.com/Crunch-io/rcrunch/blob/49cf2526c0c54d05550b6401e0b97a0beeaa1640/inst/app.crunch.io/api/datasets/1.json#L34) from the `crunch` package, an extra, nonsense attribute was added to the JSON just to ensure that the code doesn't break if there are new, unknown features added to the API response. That way, if the API grows new features, people who are using your package don't get errors if they haven't upgraded to the latest release that recognizes the new feature.

If you're responsible for the API as well as the R client code that communicates with it, the plain-text mocks can be a valuable source of documentation. Indeed, the file-system tree view of the mock files gives a visual representation of your API. For example, in the [crunch](https://crunch.io/r/crunch/) package, the mocks show an API of catalogs that contain entities that may contain other subdocuments:

```
app.crunch.io/
├── api
│   ├── accounts
│   │   ├── account1
│   │   │   └── users.json
│   │   └── account1.json
│   ├── datasets
│   │   ├── 1
│   │   │   ├── export.json
│   │   │   ├── filters
│   │   │   │   └── filter1.json
│   │   │   ├── filters.json
│   │   │   ├── permissions.json
│   │   │   ├── summary-73a614.json
│   │   │   ├── variables
│   │   │   │   ├── birthyr
│   │   │   │   │   ├── summary-73a614.json
│   │   │   │   │   └── values-3d4982.json
│   │   │   │   ├── birthyr.json
│   │   │   │   ├── gender
│   │   │   │   │   ├── summary.json
│   │   │   │   │   └── values-51980f.json
│   │   │   │   ├── gender.json
│   │   │   │   ├── starttime
│   │   │   │   │   └── values-3d4982.json
│   │   │   │   ├── starttime.json
│   │   │   │   ├── textVar
│   │   │   │   │   └── values-641ef3.json
│   │   │   │   ├── textVar.json
│   │   │   │   └── weights.json
│   │   │   ├── variables-d118fa.json
│   │   │   └── variables.json
│   │   ├── 1.json
│   │   └── search-c89aba.json
│   └── users.json
└── api.json
```

---
title: "httptest: A Test Environment for HTTP Requests in R"
description: "This vignette covers the core features of the httptest package, focusing on how to mock HTTP responses, how to make assertions about requests, and how to record real requests for future use as mocks."
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{httptest: A Test Environment for HTTP Requests in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, results='hide', echo=FALSE, message=FALSE}
options(width=120)
```

Testing code and packages that communicate with remote servers can be painful. Dealing with authentication, bootstrapping server state, cleaning up objects that may get created during the test run, network flakiness, and other complications can make testing seem too costly to bother with. But it doesn't need to be that hard. The `httptest2` package lets you test R code that constructs API requests and handles their responses, all without requiring access to the remote service during the test run. This makes tests easy to write and fast to run.

This vignette covers some of the core features of the `httptest2` package, focusing on how to mock HTTP responses, how to make other assertions about requests, and how to record real requests for future use as mocks. Note that `httptest2` requires the `testthat` package, and it follows the testing conventions and interfaces defined there, extending them with some additional wrappers and expectations. If you're not familiar with `testthat`, see the ["Testing" chapter](https://r-pkgs.org/tests.html) chapter of Hadley Wickham's _R Packages_ book.

`httptest2` is designed for use with packages that rely on the [`httr2`](https://httr2.r-lib.org/) requests library—it is a bridge between `httr2` and `testthat`. If you're using [`httr`](https://httr.r-lib.org/), use [`httptest`](https://enpiar.com/r/httptest/), on which this package is based.

## Getting started

First, we need to add `httptest2` to our package test suite. Once you've configured your package to use `testthat` (`usethis::use_testthat()` is one way), run `httptest2::use_httptest2()`. This adds `httptest2` to your package `DESCRIPTION` and makes sure it is loaded in your test setup code.

`httptest2` provides several contexts, which are "with"-style functions that you wrap around other code you want to execute.

* `without_internet()`: any `httr2` request performed will raise an informative error
* `with_mock_api()`: requests will look for an associated mock file to return instead of making a real request; if none found, it will raise an informative error
* `capture_requests()`: real requests will be saved to mock files so that `with_mock_api()` can load them
* `with_mock_dir()`: wrapper around `capture_requests()` and `with_mock_api()`, switching between them based on the existence of the given directory. If it exists, `with_mock_api()` will use it; if not, real requests will be made and responses will be recorded into it so that subsequent runs will use them as mocks. To re-record real responses, delete the directory.

## Basic tests from a live server

A good place to start is with a small number of end-to-end tests: things that you can run against the real API to confirm that the whole workflow functions correctly. To do this, we'll use the `with_mock_dir()` context.

Let's suppose we're making a package from the first example from the `httr2` vignette on [wrapping APIs](https://httr2.r-lib.org/articles/wrapping-apis.html), which wraps the [faker API](https://fakerapi.it/en):

```R
faker_person <- function(gender = NULL, birthday_start = NULL, birthday_end = NULL, quantity = 1, locale = "en_US", seed = NULL) {
  faker(
    "persons",
    gender = gender,
    birthday_start = birthday_start,
    birthday_end = birthday_end,
    quantity = quantity,
    locale = locale,
    seed = seed
  )
}

faker <- function(resource, ..., quantity = 1, locale = "en_US", seed = NULL) {
  params <- list(
    ...,
    quantity = quantity,
    locale = locale,
    seed = seed
  )
  names(params) <- paste0("_", names(params))

  request("https://fakerapi.it/api/v1") %>%
    req_url_path_append(resource) %>%
    req_url_query(!!!params) %>%
    req_user_agent("my_package_name (http://my.package.web.site)") %>%
    req_perform() %>%
    resp_body_json()
}
```

We could add a test file at `tests/testthat/test-person.R` in our package and put this test in it:

```r
test_that("We can get people", {
  expect_length(faker_person("female", quantity = 2), 2)
})
```

If we run that test, it will make a live API call and return a response containing 2 fake people. Every time we run it, it will make an API request, and if there is no network connection, or the API happens to be down, the test will fail.

Instead, let's wrap this test in `with_mock_dir()` so that we can record that response the first time we run it and then load it as a mock every time after.

```r
with_mock_dir("person", {
  test_that("We can get people", {
    expect_length(faker_person("female", quantity = 2), 2)
  })
})
```

To illustrate how that works, let's run that again outside of the test and with some verbose messages turned on.

```r
options(httptest2.verbose = TRUE)

with_mock_dir("person", {
  system.time(peeps <- faker_person("female", quantity = 2))
})

## Recording responses to tests/testthat/person
## Writing /Users/you/fakerpkg/tests/testthat/person/fakerapi.it/api/v1/persons-fdcc43.json
##    user  system elapsed
##   0.022   0.003   0.611
```

Because `tests/testthat/person` does not exist, it makes a real request, which takes around 600 milliseconds. And now if we look at the file system, we see a new file

```
tests/testthat/person/fakerapi.it/api/v1/persons-fdcc43.json
```

which contains the JSON response body from our request:

```
{
    "status": "OK",
    "code": 200,
    "total": 2,
    "data": [
        {
            "firstname": "Kyra",
            "lastname": "Schuster",
            "email": "mraz.jeffery@abernathy.com",
            "phone": "+5531037042476",
            "birthday": "1932-09-11",
            "gender": "female",
            ...
        },
        ...
    ]
}
```

Note how the file path matches the request URL. Requests are translated to mock file paths according to several rules that incorporate the request method, URL, query parameters, and body.
First, the request protocol is removed from the URL.
Second, if the request URL contains a query string, it will be popped off,
hashed by `digest::digest()`, and the first six characters appended to the
file being read. Third, request bodies are similarly hashed and
appended. Finally, if a request method other than GET is used it will be
appended to the end of the end of the file name.

If you want to know what request was made, you can run the code inside `without_internet()`, and the error message will tell you:

```r
without_internet({
  faker_person("female", quantity = 2)
})

## Error: An unexpected request was made:
## GET https://fakerapi.it/api/v1/persons?_gender=female&_quantity=2&_locale=en_US
## Run `rlang::last_error()` to see where the error occurred.
```

Back to `with_mock_dir()`: the next time we run the same code, we'll see that it loads the mock file.

```r
with_mock_dir("person", {
  system.time(peeps2 <- faker_person("female", quantity = 2))
})

## Using mocks found in tests/testthat/person
## Reading /Users/you/fakerpkg/tests/testthat/person/fakerapi.it/api/v1/persons-fdcc43.json
##    user  system elapsed
##   0.011   0.000   0.011
```

Because the `person` directory exists now, `with_mock_dir()` uses the previously recorded mocks, matches our exact query with the one we did before, and returns it instantly, without making a network connection. We can see that it is using the previously recorded response because the result is identical---with the faker API, you get different data back for every request.

```r
identical(peeps, peeps2)

## [1] TRUE
```

Note that none of the test code inside `with_mock_dir()` looks any different from how you'd write it if you were testing against a live server using just `testthat`. The goal is to make your tests just as natural to write as if you were using your package normally.

> Do your recorded responses contain sensitive information? Are they unnecessarily large? See `vignette("redacting", package = "httptest2")` for how to trim and sanitize recorded mocks.

## Going deeper

`with_mock_dir()` is convenient to use, but you probably don't want all of your tests to be based on real requests. Among the reasons:

* Some API responses contain way more content than is necessary to test your R code around them: 100 records when 2 will suffice, include metadata that you don't care about and can't meaningfully assert things about, and so on. In the interest of minimally reproducible examples, and of making tests readable, it often makes sense to take an actual API response and delete a lot of its content, or even to fabricate one entirely.
* It's good to keep an API mock fixed so you know exactly what is in it. If you re-recorded a Twitter API response of, say, the most recent 10 tweets with `#rstats`, the specific content will change every time you record it, so your tests can't say much about what is in the response without having to rewrite them every time too.
* Some conditions (rate limiting, server errors, etc.) are difficult to test with real responses, but if you can manually create a API mock with, say, a 503 response status code and test how your code handles it, you can have confidence of how your package will respond when that rare event happens with the real API.
* Re-recording all responses can make for a huge code diff, which can blow up your repository size and make code review harder.

Moreover, we generally don't need to test what the server does with our requests. We're responsible for testing the R code on either side: (1) given some inputs, does my code make the correct HTTP request(s); and (2) does my code correctly handle the types of responses that the server can return? We often don't need to make real requests to test that.

Sometimes it is more clear what you're testing if you focus only on the requests. One case is when the response itself isn't that interesting or doesn't tell you that the request did the correct thing. For example, if you're testing a POST request that alters the state of something on the server and returns 204 No Content status on success, nothing in the response itself (which would be stored in the mock file) tells you that the request you made was shaped correctly—the response has no content. A more transparent, readable test would just assert that the POST request was made to the right URL and had the expected request body.

Another case where you might want to focus just on the request, without recording real API responses, is when you don't actually want to make the requests. Perhaps you're testing a function that would call `DELETE` on some resource. Or, to make it concrete, suppose you were testing a function that sends a tweet: you probably don't want to send a real test tweet out into the world.

Here's a function that would send a tweet using the Twitter API. Consulting the [API docs](https://developer.twitter.com/en/docs/twitter-api/tweets/manage-tweets/api-reference/post-tweets), it looks like this is the request we'd make:

```r
send_tweet <- function(text) {
  resp <- request("https://api.twitter.com/2/tweets") %>%
    req_method("POST") %>%
    req_body_json(list(text = text)) %>%
    req_perform()

  resp_body_json(resp)$data
}
```

Of course, this is overly simplified. According to the docs, there are lots more arguments you would want to add, and also your real function would need to apply your OAuth token. If you were really writing this function (or broader package), you'd solve that. For the purposes of illustration here, let's assume that's taken care of.

How do we test this without making a real request? First, let's run the function inside `without_internet()` so we can see what request it makes:

```r
without_internet({
  send_tweet("Hello world!")
})

## Error: An unexpected request was made:
## POST https://api.twitter.com/2/tweets {"text":"Hello world!"}
## Run `rlang::last_error()` to see where the error occurred.
```

That looks right according to the API documentation. We can now turn this into a test that asserts we're making the correct request, using `expect_POST()`.

```r
without_internet({
  test_that("We can send a tweet", {
    expect_POST(
      send_tweet("Hello world!"),
      'https://api.twitter.com/2/tweets',
      '{"text":"Hello world!"}'
    )
  })
})

## Test passed 🌈
```

`expect_POST()` (and the other `expect_VERB` functions) look for the request method, URL, and body, just as they are printed in the error message if the request is not caught by `expect_POST()`.

If the test fails because the request that is made doesn't match what we expected, it looks like this:

```r
without_internet({
  test_that("We can send a tweet", {
    expect_POST(
      send_tweet("Goodbye world!"),
      'https://api.twitter.com/2/tweets',
      '{"text":"Hello world!"}'
    )
  })
})

## ── Failure (Line 3): We can send a tweet ───────────────────────────────────────
## An unexpected request was made:
##   Actual:   POST https://api.twitter.com/2/tweets {"text":"Goodbye world!"}
##   Expected: POST https://api.twitter.com/2/tweets {"text":"Hello world!"}
```

Next, suppose we wanted to test how our code handles the API response. In our example, we aren't doing much, just extracting the "data" element from the response, but in practice there could be much more involved code to turn API responses into objects that make intuitive sense to R users.

We can still set up this test without making a real request by creating a mock based on the API documentation, which includes a sample response. To start, let's switch from `without_internet()` to `with_mock_api()` and re-run the function:

```r
with_mock_api({
  send_tweet("Hello world!")
})

## Error: An unexpected request was made:
## POST https://api.twitter.com/2/tweets {"text":"Hello world!"}
## Expected mock file: api.twitter.com/2/tweets-8d6065-POST.*
## Run `rlang::last_error()` to see where the error occurred.
```

We got the same error message before, but there is an extra line telling us what mock file it was looking to load. If we take the JSON example response from the API documentation and put it in that location, `with_mock_api()` will load it.

```r
dir.create("tests/testthat/api.twitter.com/2", recursive = TRUE)
cat('{
  "data": {
    "id": "1445880548472328192",
    "text": "Hello world!"
  }
}', file = "tests/testthat/api.twitter.com/2/tweets-8d6065-POST.json")

with_mock_api({
  send_tweet("Hello world!")
})

## $id
## [1] "1445880548472328192"
##
## $text
## [1] "Hello world!"
##
```

Now we can test that we can make that request and the mock response is loaded:

```r
with_mock_api({
  test_that("We can send a tweet", {
    expect_equal(
      send_tweet("Hello world!")$text,
      "Hello world!"
    )
  })
})

## Test passed 🥇
```

Because requests made in `with_mock_api()` that don't find a corresponding mock file raise the same error as `without_internet()`, we can combine mock tests and tests that assert that a request would be made:

```r
with_mock_api({
  test_that("We can send a tweet", {
    # This one has a mock file now
    expect_equal(
      send_tweet("Hello world!")$text,
      "Hello world!"
    )
    # This one does not
    expect_POST(
      send_tweet("Goodbye world!"),
      'https://api.twitter.com/2/tweets',
      '{"text":"Goodbye world!"}'
    )
  })
})

## Test passed 🌈
```

For more examples of cases to test like this, see the [`httptest` vignette](https://enpiar.com/r/httptest/articles/httptest.html) and [this blog post](https://enpiar.com/2017/06/21/7-hard-testing-problems-made-easy-by-httptest/).

# expect_request without_internet

    Code
      expect_POST(req_perform(this_req), "http://httpbin.not/get", "{\"test\":false}")
    Condition
      Error:
      ! An unexpected request was made:
        Actual:   POST http://httpbin.not/get {"test":true}
        Expected: POST http://httpbin.not/get {"test":false}

# expect_request_header with mock API

    Code
      expect_request_header(request("https://test.api/object1/") %>% req_headers(
        Accept = "image/png") %>% req_perform(), accept = "image/jpeg")
    Condition
      Error:
      ! Expected Header "accept" to match regexp "image/jpeg".
      Actual text:
      x | image/png

---

    Code
      expect_request_header(request("https://test.api/object1/") %>% req_headers(
        Accept = "image/png") %>% req_perform(), accept = NULL)
    Condition
      Error:
      ! Expected Header "accept" to be NULL.
      Differences:
      `actual` is a character vector ('image/png')
      `expected` is NULL

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-api.R
\name{with_mock_api}
\alias{with_mock_api}
\alias{use_mock_api}
\alias{stop_mocking}
\title{Serve a mock API from files}
\usage{
with_mock_api(expr)

use_mock_api()

stop_mocking()
}
\arguments{
\item{expr}{Code to run inside the mock context}
}
\value{
\code{with_mock_api()} returns the result of \code{expr}. \code{use_mock_api()} and
\code{stop_mocking()} return nothing.
}
\description{
In this context, HTTP requests attempt to load API response fixtures from
files. This allows test code to proceed evaluating code that expects
HTTP requests to return meaningful responses. Requests that do not have a
corresponding fixture file raise errors, like how \code{\link[=without_internet]{without_internet()}}
does.
}
\details{
\code{use_mock_api()} and \code{stop_mocking()} allow you to turn on/off request
mocking for more convenient use in an interactive session.

Requests are translated to mock file paths according to several rules that
incorporate the request method, URL, query parameters, and body. See
\code{\link[=build_mock_url]{build_mock_url()}} for details.

File paths for API fixture files may be relative to the 'tests/testthat'
directory, i.e. relative to the .R test files themselves. This is the default
location for storing and retrieving mocks, but you can put them anywhere you
want as long as you set the appropriate location with \code{\link[=.mockPaths]{.mockPaths()}}.
}
\examples{
library(httr2)
with_mock_api({
  # There are no mocks recorded in this example, so catch this request with
  # expect_GET()
  expect_GET(
    request("https://cran.r-project.org") \%>\% req_perform(),
    "https://cran.r-project.org"
  )
  # For examples with mocks, see the tests and vignettes
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/use-httptest2.R
\name{use_httptest2}
\alias{use_httptest2}
\title{Use 'httptest2' in your tests}
\usage{
use_httptest2(path = ".")
}
\arguments{
\item{path}{character path to the package}
}
\value{
Nothing: called for file system side effects.
}
\description{
This function adds \code{httptest2} to Suggests in the package DESCRIPTION and
loads it in \code{tests/testthat/setup.R}. Call it once when you're setting up
a new package test suite.
}
\details{
The function is idempotent: if \code{httptest2} is already added to these files,
no additional changes will be made.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignette.R
\name{start_vignette}
\alias{start_vignette}
\alias{end_vignette}
\title{Set mocking/capturing state for a vignette}
\usage{
start_vignette(dir, ...)

end_vignette()
}
\arguments{
\item{dir}{Root file path for the mocks for this vignette. A good idea is
to use the file name of the vignette itself.}

\item{...}{Optional arguments passed to \code{start_capturing()}}
}
\value{
Nothing; called for its side effect of starting/ending
response recording or mocking.
}
\description{
Use \code{start_vignette()} to either use previously recorded responses, if they
exist, or capture real responses for future use.
}
\details{
In a vignette or other R Markdown or Sweave document, place
\code{start_vignette()} in an R code block at the beginning,
before the first API request is made, and put
\code{end_vignette()} in a R code chunk at the end. You may
want to make those R code chunks have \code{echo=FALSE} in order to hide the fact
that you're calling them.

As in \code{\link[=with_mock_dir]{with_mock_dir()}}, the behavior changes based on the existence of the \code{dir} directory. The
first time you build the vignette, the directory won't exist yet, so it will
make real requests and record them inside of \code{dir}. On subsequent runs, the
mocks will be used. To record fresh responses from the server, delete the
\code{dir} directory, and the responses will be recorded again the next time the
vignette runs.

If you have additional setup code that you'd like available across all of
your package's vignettes, put it in \code{inst/httptest2/start-vignette.R} in your
package, and it will be called in \code{start_vignette()} before the mock/record
context is set. Similarly, teardown code can go in
\code{inst/httptest2/end-vignette.R}, evaluated in \code{end_vignette()} after mocking
is stopped.
}
\seealso{
\code{\link[=start_capturing]{start_capturing()}} for how requests are recorded; \code{\link[=use_mock_api]{use_mock_api()}}
for how previously recorded requests are loaded; \code{\link[=change_state]{change_state()}} for how to
handle recorded requests when the server state is changing;
\code{vignette("vignettes", package = "httptest2")} for an overview of all
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-paths.R
\name{.mockPaths}
\alias{.mockPaths}
\title{Set an alternate directory for mock API fixtures}
\usage{
.mockPaths(new)
}
\arguments{
\item{new}{Either a character vector of path(s) to add, or \code{NULL} to reset
to the default.}
}
\value{
If \code{new} is omitted, the function returns the current search paths, a
a character vector. If \code{new} is provided, the updated value will be returned
invisibly.
}
\description{
By default, \code{with_mock_api()} will look for and \code{capture_requests()} will
write mocks to your package's \code{tests/testthat} directory, or else the current
working directory if that path does not exist. If you want to look in or
write to other places, call \code{.mockPaths()} to add directories to the search
path.
}
\details{
It works like \code{\link[base:libPaths]{base::.libPaths()}}: any directories you specify will be added
to the list and searched first. The default directory will be searched last.
Only unique values are kept: if you provide a path that is already found in
\code{.mockPaths()}, the result effectively moves that path to the first position.

For finer-grained control, or to completely override the default behavior
of searching in the current working directory, you can set the option
"httptest2.mock.paths" directly.
}
\examples{
.mockPaths()
.mockPaths("/var/somewhere/else")
.mockPaths()
.mockPaths(NULL)
.mockPaths()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture-requests.R
\name{save_response}
\alias{save_response}
\title{Write out a captured response}
\usage{
save_response(response, file, simplify = TRUE)
}
\arguments{
\item{response}{An \code{httr2_response} object}

\item{file}{String file path to write to}

\item{simplify}{logical: if \code{TRUE} (default), JSON responses with status 200
and a supported \code{Content-Type}
will be written as just the text of the response body. In all other cases,
and when \code{simplify} is \code{FALSE}, the \code{httr2_response} object will be written
out to a .R file using \code{\link[base:dput]{base::dput()}}.}
}
\value{
The character file name that was written out
}
\description{
Write out a captured response
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redactor.R
\name{set_redactor}
\alias{set_redactor}
\title{Set a response redactor}
\usage{
set_redactor(FUN)
}
\arguments{
\item{FUN}{A function or expression that modifies \code{httr2_response} objects.
Specifically, a valid input is one of:
\itemize{
\item A function taking a single argument, the \code{httr2_response},
and returning a valid \code{httr2_response} object.
\item A formula as shorthand for an anonymous function with \code{.} as the
"response" argument, as in the \code{purrr} package. That is, instead of
\code{function (response) redact_headers(response, "X-Custom-Header")}, you can
use \code{~ redact_headers(., "X-Custom-Header")}
\item A list of redacting functions/formulas, which will be executed
in sequence on the response
\item \code{NULL}, to override the default \code{redact_cookies()}.
}}
}
\value{
Invisibly, the redacting function, validated and perhaps modified.
Formulas and function lists are turned into proper functions. \code{NULL} as input
returns the \code{force()} function.
}
\description{
A redactor is a function that alters the response content being written
out in the \code{\link[=capture_requests]{capture_requests()}} context, allowing you to remove sensitive
values, such as authentication tokens,
as well as any other modification or truncation of the response body. By
default, the \code{\link[=redact_cookies]{redact_cookies()}} function will be used to purge standard
auth methods, but \code{set_redactor()} allows you to provide a different one.
}
\details{
Alternatively, you can put a redacting function in \code{inst/httptest2/redact.R}
in your package, and
any time your package is loaded (as in when running tests or building
vignettes), the function will be used automatically.
}
\examples{
# Shorten UUIDs in response body/URLs to their first 6 digits:
set_redactor(function(resp) gsub_response(resp, "([0-9a-f]{6}[0-9a-f]{26}", "\\\\1"))
# Restore the default
set_redactor(redact_cookies)
}
\seealso{
For further details on how to redact responses, see \code{vignette("redacting", package = "httptest2")}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redact.R
\name{gsub_response}
\alias{gsub_response}
\title{Find and replace within a response object}
\usage{
gsub_response(response, pattern, replacement, ...)
}
\arguments{
\item{response}{An \code{httr2_response} or \code{http2_request} object to sanitize.}

\item{pattern}{From \code{\link[base:grep]{base::gsub()}}: "character string containing a regular
expression (or character string for \code{fixed = TRUE}) to be matched in the
given character vector." Passed to \code{gsub()}. See the docs for \code{gsub()} for
further details.}

\item{replacement}{A replacement for the matched pattern, possibly including
regular expression backreferences. Passed to \code{gsub()}. See the docs for
\code{gsub()} for further details.}

\item{...}{Additional logical arguments passed to \code{gsub()}: \code{ignore.case},
\code{perl}, \code{fixed}, and \code{useBytes} are the possible options.}
}
\value{
An \code{httr2_response} object, same as was passed in, with the
pattern replaced in the URLs and bodies.
}
\description{
This function passes its arguments to \code{\link[base:grep]{base::gsub()}} in order to find and
replace string patterns (regular expressions) within
the URL and the response body of \code{httr2_response} objects.
}
\details{
Note that, unlike \code{gsub()}, the first argument of the function is \code{response},
not \code{pattern}, while the equivalent argument in \code{gsub()}, "\code{x}", is placed
third. This difference is to maintain consistency with the other redactor
functions in \code{httptest2}, which all take \code{response} as the first argument.

This function also can be applied to an \code{http2_request} object to replace
patterns inside the request URL.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redact.R
\name{as.redactor}
\alias{as.redactor}
\title{Wrap a redacting expression as a proper function}
\usage{
as.redactor(fmla)
}
\arguments{
\item{fmla}{Partial expression (a \code{formula}) to turn into a function}
}
\value{
A \code{function}.
}
\description{
Redactors take a \code{httr2_response} or \code{httr2_request} as their first argument, and some take additional
arguments: \code{redact_headers()}, for example, requires that you specify
\code{headers}. This function allows you to take a simplified expression via a
formula, similar to what \code{purrr} does, so that you can provide the function
to \code{capture_requests()}.
}
\details{
For example, \code{as.redactor(~ redact_headers(., "X-Custom-Header"))} is equivalent
to \code{function (response) redact_headers(response, "X-Custom-Header")}. This
allows you to do
\code{set_redactor(~ redact_headers(., "X-Custom-Header"))}.
}
\seealso{
\code{\link[=capture_requests]{capture_requests()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vignette.R
\name{change_state}
\alias{change_state}
\title{Handle a change of server state}
\usage{
change_state()
}
\value{
Invisibly, the return of \code{.mockPaths()} with the new path added.
}
\description{
In a vignette, put a call to \code{change_state()} before any code block that
makes a change on
the server, or rather, before any code block that might repeat the same
request previously done and expect a different result.
}
\details{
\code{change_state()} works by layering a new directory on top of the existing
\code{\link[=.mockPaths]{.mockPaths()}}, so fixtures are recorded/loaded there, masking rather than
overwriting previously recorded responses for the same request. In
vignettes, these mock layers are subdirectories with integer names.
}
\seealso{
\code{\link[=start_vignette]{start_vignette()}};
\code{vignette("vignettes", package = "httptest2")} for an overview of all
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/without-internet.R
\name{without_internet}
\alias{without_internet}
\alias{block_requests}
\title{Make all HTTP requests raise an error}
\usage{
without_internet(expr)

block_requests()
}
\arguments{
\item{expr}{Code to run inside the mock context}
}
\value{
The result of \code{expr}
}
\description{
\code{without_internet()} simulates the situation when any network request will
fail, as in when you are without an internet connection. Any HTTP request
through \code{httr2} will raise an error.
}
\details{
\code{block_requests()} and \code{stop_mocking()} allow you to turn on/off request
blocking for more convenient use in an interactive session.

The error message raised has a well-defined shape, made of three
elements, separated by space: (1) the request
method (e.g. "GET"); (2) the request URL; and
(3) the request body, if present. The verb-expectation functions,
such as \code{\link[=expect_GET]{expect_GET()}} and \code{\link[=expect_POST]{expect_POST()}}, look for this shape.
}
\examples{
library(httr2)
library(testthat, warn.conflicts = FALSE)
without_internet({
  expect_error(
    request("http://httpbin.org/get") \%>\% req_perform(),
    "GET http://httpbin.org/get"
  )
  expect_error(
    request("http://httpbin.org/put") \%>\%
      req_method("PUT") \%>\%
      req_body_json(list(a = 1)) \%>\%
      req_perform(),
    'PUT http://httpbin.org/put {"a":1}',
    fixed = TRUE
  )
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/capture-requests.R
\name{capture_requests}
\alias{capture_requests}
\alias{start_capturing}
\alias{stop_capturing}
\title{Record API responses as mock files}
\usage{
capture_requests(expr, simplify = TRUE)

start_capturing(simplify = TRUE)

stop_capturing()
}
\arguments{
\item{expr}{Code to run inside the context}

\item{simplify}{logical: if \code{TRUE} (default), plain-text responses with status 200
will be written as just the text of the response body. In all other cases,
and when \code{simplify} is \code{FALSE}, the \code{httr2_response} object will be written out to
a .R file using \code{\link[base:dput]{base::dput()}}.}
}
\value{
\code{capture_requests()} returns the result of \code{expr}. \code{start_capturing()}
invisibly returns the destination directory.
\code{stop_capturing()} returns nothing; it is called for its side effects.
}
\description{
\code{capture_requests()} is a context that collects the responses from requests
you make and stores them as mock files. This enables you to perform a series
of requests against a live server once and then build your test suite using
those mocks, running your tests in \code{\link[=with_mock_api]{with_mock_api()}}.
}
\details{
\code{start_capturing()} and \code{stop_capturing()} allow you to turn on/off request
recording for more convenient use in an interactive session.

Recorded responses are written out as plain-text files. By storing fixtures
as plain-text files, you can
more easily confirm that your mocks look correct, and you can more easily
maintain them without having to re-record them. If the API changes subtly,
such as when adding an additional attribute to an object, you can just touch
up the mocks.

If the response has status \verb{200 OK} and the \code{Content-Type}
maps to a supported file extension---currently \code{.json},
\code{.html}, \code{.xml}, \code{.txt}, \code{.csv}, and \code{.tsv}---just the response body will be
written out, using the appropriate extension. \verb{204 No Content} status
responses will be stored as an empty file with extension \code{.204}. Otherwise,
the response will be written as a \code{.R} file containing syntax that, when
executed, recreates the \code{httr2_response} object.

Files are saved to the first directory in \code{\link[=.mockPaths]{.mockPaths()}}, which if not
otherwise specified is either "tests/testthat" if it exists
(as it should if you are in the root directory of your package),
else the current working directory.
If you have trouble when recording responses, or are unsure where the files
are being written, set \code{options(httptest2.verbose = TRUE)} to print a message
for every file that is written containing the absolute path of the file.
}
\examples{
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Setup so that our examples clean up after themselves
tmp <- tempfile()
.mockPaths(tmp)
on.exit(unlink(tmp, recursive = TRUE))

library(httr2)
capture_requests({
  request("http://httpbin.org/get") \%>\% req_perform()
  request("http://httpbin.org/response-headers") \%>\%
    req_headers(`Content-Type` = "application/json") \%>\%
    req_perform()
})
# Or:
start_capturing()
request("http://httpbin.org/get") \%>\% req_perform()
request("http://httpbin.org/response-headers") \%>\%
  req_headers(`Content-Type` = "application/json") \%>\%
  req_perform()
stop_capturing()
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=build_mock_url]{build_mock_url()}} for how requests are translated to file paths.
And see \code{vignette("redacting", package = "httptest2")}
for details on how to prune sensitive content from responses when recording.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-request.R
\name{expect_verb}
\alias{expect_verb}
\alias{expect_GET}
\alias{expect_POST}
\alias{expect_PUT}
\alias{expect_PATCH}
\alias{expect_DELETE}
\alias{expect_no_request}
\title{Expectations for mocked HTTP requests}
\usage{
expect_GET(object, url = "", ...)

expect_POST(object, url = "", ...)

expect_PATCH(object, url = "", ...)

expect_PUT(object, url = "", ...)

expect_DELETE(object, url = "", ...)

expect_no_request(object, ...)
}
\arguments{
\item{object}{Code to execute that may cause an HTTP request}

\item{url}{character: the URL you expect a request to be made to. Default is
an empty string, meaning that you can just assert that a request is made with
a certain method without asserting anything further.}

\item{...}{character segments of a request payload you expect to be included
in the request body, to be joined together by \code{paste0()}. You may also pass
any of the following named logical arguments, which will be passed to
\code{\link[base:grep]{base::grepl()}}:
\itemize{
\item \code{fixed}: Should matching take the pattern as is or treat it as a regular
expression. Default: \code{TRUE}, and note that this default is the opposite of
the default in \code{grepl}. (The rest of the arguments follow its defaults.)
\item \code{ignore.case}: Should matching be done case insensitively? Default:
\code{FALSE}, meaning matches are case sensitive.
\item \code{perl}: Should Perl-compatible regular expressions be used? Default: \code{FALSE}
\item \code{useBytes}: Should matching be done byte-by-byte rather than
character-by-character? Default: \code{FALSE}
}}
}
\value{
A \code{testthat} 'expectation'.
}
\description{
The mock contexts in \code{httptest2} can raise errors or messages when requests
are made, and those (error) messages have three
elements, separated by space: (1) the request
method (e.g. "GET"); (2) the request URL; and
(3) the request body, if present.
These verb-expectation functions look for this message shape. \code{expect_PUT},
for instance, looks for a request message that starts with "PUT".
}
\examples{
library(httr2)
without_internet({
  expect_GET(
    request("http://httpbin.org/get") \%>\% req_perform(),
    "http://httpbin.org/get"
  )
  expect_GET(
    request("http://httpbin.org/get") \%>\% req_perform(),
    "http://httpbin.org/[a-z]+",
    fixed = FALSE # For regular expression matching
  )
  expect_PUT(
    request("http://httpbin.org/put") \%>\%
      req_method("PUT") \%>\%
      req_body_json(list(a = 1)) \%>\%
      req_perform(),
    "http://httpbin.org/put",
    '{"a":1}'
  )
  # Don't need to assert the request body, or even the URL
  expect_PUT(
    request("http://httpbin.org/put") \%>\%
      req_method("PUT") \%>\%
      req_body_json(list(a = 1)) \%>\%
      req_perform()
  )
  expect_no_request(rnorm(5))
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redactor.R
\name{get_current_redactor}
\alias{get_current_redactor}
\title{Fetch the active redacting function}
\usage{
get_current_redactor()
}
\value{
A redacting function.
}
\description{
Called inside \code{\link[=capture_requests]{capture_requests()}}. If using the default redactor, it checks
each time it is called to see if any new packages have been attached, in case
there are package redactors in them.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-api.R
\name{find_mock_file}
\alias{find_mock_file}
\title{Go through mock paths to find the local mock file location}
\usage{
find_mock_file(file)
}
\arguments{
\item{file}{A file path, as generated by \code{\link[=build_mock_url]{build_mock_url()}}.}
}
\value{
A path to a file that exists, or \code{NULL} if none found.
}
\description{
Go through mock paths to find the local mock file location
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-mock-url.R
\name{build_mock_url}
\alias{build_mock_url}
\title{Convert a request to a mock file path}
\usage{
build_mock_url(req)
}
\arguments{
\item{req}{A \code{httr2_request} object}
}
\value{
A file path and name, without an extension. The file, or a file with
some extension appended, may or may not
exist: existence is not a concern of this function.
}
\description{
Requests are translated to mock file paths according to several rules that
incorporate the request method, URL, query parameters, and body.
}
\details{
First, the request protocol, such as "https://", is removed from the URL.
Second, if the request URL contains a query string, it will be popped off,
hashed by \code{\link[digest:digest]{digest::digest()}}, and the first six characters appended to the
file being read. Third, request bodies are similarly hashed and
appended. Finally, if a request method other than GET is used it will be
appended to the end of the end of the file name.

Mock file paths also have a file extension appended, based on the
\code{Content-Type} of the response, though this function, which is only concerned
with the request, does not add the extension. In an
HTTP API, a "directory" itself is a resource,
so the extension allows distinguishing directories and files in the file
system. That is, a mocked \verb{GET http://example.com/api/} may read a
"example.com/api.json" file, while
\verb{GET http://example.com/api/object1/} reads "example.com/api/object1.json".

Other examples:
\itemize{
\item \verb{GET http://example.com/api/object1/?a=1} may read
"example.com/api/object1-b64371.xml".
\item \verb{POST http://example.com/api/object1/?a=1} may read
"example.com/api/object1-b64371-POST.json".
}

Note that if you are trying to guess the mock file paths corresponding to a
test for which you intend to create a mock file manually,
instead of trying to build the URL, you should run the test
with \code{with_mock_api()} as the error message will contain the mock file path.
}
\seealso{
\code{\link[=with_mock_api]{with_mock_api()}} \code{\link[=capture_requests]{capture_requests()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trace.R
\name{with_trace}
\alias{with_trace}
\title{Wrapper around 'trace' to untrace when finished}
\usage{
with_trace(
  x,
  where = topenv(parent.frame()),
  print = getOption("httptest2.debug.trace", FALSE),
  ...,
  expr
)
}
\arguments{
\item{x}{Name of function to trace. See \code{\link[base:trace]{base::trace()}}.}

\item{where}{where to look for the function to be traced.}

\item{print}{Logical: print a message every time the traced function is hit?
Default is \code{FALSE}; note that in \code{trace}, the default is \code{TRUE}.}

\item{...}{Additional arguments to pass to \code{trace}. At minimum, must include
either \code{tracer} or \code{exit}.}

\item{expr}{Code to run inside the context}
}
\value{
The result of \code{expr}
}
\description{
Wrapper around 'trace' to untrace when finished
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mock-dir.R
\name{with_mock_dir}
\alias{with_mock_dir}
\title{Use or create mock files depending on their existence}
\usage{
with_mock_dir(dir, expr, simplify = TRUE, replace = TRUE)
}
\arguments{
\item{dir}{character string, unique folder name that will be used or created
under \verb{tests/testthat/}}

\item{expr}{Code to run inside the context}

\item{simplify}{logical: if \code{TRUE} (default), plain-text responses with status 200
will be written as just the text of the response body. In all other cases,
and when \code{simplify} is \code{FALSE}, the \code{httr2_response} object will be written out to
a .R file using \code{\link[base:dput]{base::dput()}}.}

\item{replace}{Logical: should \code{dir} replace the contents of \code{.mockPaths()}
(default) or be added in front of the existing paths? The default behavior
here is the opposite of \code{.mockPaths()} so that the tests inside of
\code{with_mock_dir()} are fully isolated.}
}
\description{
This context will switch the \code{\link[=.mockPaths]{.mockPaths()}} to \code{tests/testthat/dir}
(and then resets it to what it was before).
If the \code{tests/testthat/dir} folder doesn't exist, \code{\link[=capture_requests]{capture_requests()}} will
be run to create mocks.
If it exists, \code{\link[=with_mock_api]{with_mock_api()}} will be run.
To re-record mock files, simply delete \code{tests/testthat/dir} and run the test.
}
\seealso{
\code{vignette("httptest2")} for usage examples.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/httptest2.R
\docType{package}
\name{httptest2}
\alias{httptest2-package}
\alias{httptest2}
\title{\code{httptest2}: A Test Environment for HTTP Requests}
\description{
The \code{httptest2} package lets you test R code that wraps an API without
requiring access to the remote service. It provides multiple test contexts
that mock the network connection in different ways. \code{\link[=with_mock_api]{with_mock_api()}} lets
you provide custom fixtures as responses to requests, stored as plain-text
files in your test directory. \code{\link[=without_internet]{without_internet()}} converts HTTP requests
into errors that print the request method, URL, and body payload, if
provided, allowing you to assert that a function call would make a
correctly-formed HTTP request or assert that a function does not make a
request (because if it did, it would raise an error in this context).
}
\details{
The package also includes \code{\link[=capture_requests]{capture_requests()}}, a context that collects the
responses from requests you make and stores them as mock files. This enables
you to perform a series of requests against a live server once and then build
your test suite using those mocks, running your tests in \code{with_mock_api()}.
For a context that combines both recording and mocking, use
\code{\link[=with_mock_dir]{with_mock_dir()}}: it records the first time it is run, and subsequent
runs use the mocks for offline testing.

\code{httptest2} also includes expectations to assert that HTTP requests
were---or were not---made. \code{\link[=expect_GET]{expect_GET()}}, \code{\link[=expect_PUT]{expect_PUT()}}, \code{\link[=expect_PATCH]{expect_PATCH()}},
\code{\link[=expect_POST]{expect_POST()}}, and \code{\link[=expect_DELETE]{expect_DELETE()}} assert that the specified HTTP request
is made within one of the test contexts. They catch the error or message
raised by the mocked HTTP service and check that the request URL and optional
body match the expectation. \code{\link[=expect_no_request]{expect_no_request()}} is the inverse of those: it
asserts that no error or message from a mocked HTTP service is raised.
\code{\link[=expect_request_header]{expect_request_header()}} asserts that an HTTP request, mocked or not,
contains a request header.

For an overview of testing with \code{httptest2}, see \code{vignette("httptest2")}.

When recording requests, by default \code{httptest2} looks for and redacts the
standard ways that auth credentials are passed in requests. This prevents you
from accidentally publishing your personal tokens. The redacting behavior is
fully customizable, either by providing a \code{function (response) {...}} to
\code{set_redactor()}, or by placing a function in your package's
\code{inst/httptest2/redact.R} that will be used automatically any time you record
requests with your package loaded. See \code{vignette("redacting")} for details.

\code{httptest2} also enables you to write package vignettes and other R Markdown
documents that communicate with a remote API. By adding as little as
\code{\link[=start_vignette]{start_vignette()}} to the beginning of your vignette, you can safely record
API responses from a live session, using your secret credentials. These API
responses are scrubbed of sensitive personal information and stored in a
subfolder in your \code{vignettes} directory. Subsequent vignette builds,
including on continuous-integration services, CRAN, and your package users'
computers, use these recorded responses, allowing the document to regenerate
without a network connection or API credentials. To record fresh API
responses, delete the subfolder of cached responses and re-run. See
\code{vignette("vignettes")} for more discussion and links to examples.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://enpiar.com/httptest2/}
  \item \url{https://github.com/nealrichardson/httptest2}
  \item Report bugs at \url{https://github.com/nealrichardson/httptest2/issues}
}

}
\author{
\strong{Maintainer}: Neal Richardson \email{neal.p.richardson@gmail.com} (\href{https://orcid.org/0009-0002-7992-3520}{ORCID})

Other contributors:
\itemize{
  \item Jonathan Keane \email{jkeane@gmail.com} [contributor]
  \item Maëlle Salmon \email{maelle.salmon@yahoo.se} (\href{https://orcid.org/0000-0002-2815-0399}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redact.R
\name{redact}
\alias{redact}
\alias{redact_cookies}
\alias{redact_headers}
\alias{within_body_text}
\title{Remove sensitive content from HTTP responses}
\usage{
redact_cookies(response)

redact_headers(response, headers = c())

within_body_text(response, FUN)
}
\arguments{
\item{response}{An \code{httr2_response} or \code{httr2_request} object to sanitize.}

\item{headers}{For \code{redact_headers()}, a character vector of header names to
sanitize.}

\item{FUN}{For \code{within_body_text()}, a function that takes as its argument a
character vector and returns a modified version of that. This function will
be applied to the text of the response's body.}
}
\value{
All redacting functions return a well-formed \code{httr2_response}
or \code{httr2_request} object.
}
\description{
When recording requests for use as test fixtures, you don't want to include
secrets like authentication tokens and personal ids. These functions provide
a means for redacting this kind of content, or anything you want, from
responses that \code{\link[=capture_requests]{capture_requests()}} saves.
}
\details{
\code{redact_cookies()} removes cookies from \code{httr2_response} objects
and is the default redactor in \code{capture_requests()}.
\code{redact_headers()} lets you target selected request and response headers for
redaction.
\code{within_body_text()} lets you manipulate the text of the response body
and manages the parsing of the raw (binary) data in the \code{httr_response} object.

Note that if you set a redacting function, it will also be applied to requests when loading mocks. This allows you to sanitize and/or shorten URLs in your mock files.
}
\seealso{
\code{vignette("redacting", package = "httptest2")} for a detailed discussion of what these functions do and how to customize them. \code{\link[=gsub_response]{gsub_response()}} is another redactor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expect-request-header.R
\name{expect_request_header}
\alias{expect_request_header}
\title{Test that an HTTP request is made with a header}
\usage{
expect_request_header(
  expr,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  perl = FALSE,
  useBytes = FALSE
)
}
\arguments{
\item{expr}{Code to evaluate}

\item{...}{Named headers to match. Values should either be a string (length-1
character), which will be passed to \code{\link[testthat:expect_match]{testthat::expect_match()}}, or \code{NULL} to
assert that the named header is not present in the request. To assert that a
header is merely present in the request, without asserting anything about its
contents, provide an empty string (\code{""}). Header names are always
case-insensitive; header values will be matched using the following
parameters:}

\item{fixed}{logical.  If \code{TRUE}, \code{pattern} is a string to be
    matched as is.  Overrides all conflicting arguments.}

\item{ignore.case}{if \code{FALSE}, the pattern matching is \emph{case
      sensitive} and if \code{TRUE}, case is ignored during matching.}

\item{perl}{logical.  Should Perl-compatible regexps be used?}

\item{useBytes}{logical.  If \code{TRUE} the matching is done
    byte-by-byte rather than character-by-character.  See
    \sQuote{Details}.}
}
\value{
The value of \code{expr} if there are no expectation failures
}
\description{
This expectation checks that HTTP headers (and potentially header values)
are present in a request. It works both in the mock HTTP contexts and on
"live" HTTP requests.
}
\examples{
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}

library(httr2)
expect_request_header(
  request("http://httpbin.org") \%>\%
    req_headers(Accept = "image/png") \%>\%
    req_perform(),
  accept = "image/png",
  `x-fake-header` = NULL
)
expect_request_header(
  request("http://httpbin.org") \%>\%
    req_headers(Accept = "image/png") \%>\%
    req_perform(),
  accept = ""
)
\dontshow{\}) # examplesIf}
}
